"""Generative models using the vLLM inference framework."""

import collections.abc as c
import contextlib
import importlib.util
import json
import logging
import re
import shutil
import typing as t
from functools import partial
from pathlib import Path
from time import sleep

import torch
from huggingface_hub import snapshot_download
from pydantic import conlist, create_model
from transformers.generation.configuration_utils import GenerationConfig
from transformers.models.auto.configuration_auto import AutoConfig
from transformers.models.auto.tokenization_auto import AutoTokenizer
from urllib3.exceptions import RequestError

from ..constants import (
    ATTENTION_BACKENDS,
    CUSTOM_STOP_TOKENS,
    GENERATION_KWARGS,
    GENERATIVE_PIPELINE_TAGS,
    MAX_CONTEXT_LENGTH,
    MAX_VLLM_LOGPROBS,
    MERGE_TAGS,
    REASONING_MAX_TOKENS,
    REASONING_TOKENS,
    VLLM_BF16_MIN_CUDA_COMPUTE_CAPABILITY,
)
from ..data_models import GenerativeModelOutput, HashableDict, ModelConfig
from ..enums import (
    BatchingPreference,
    GenerativeType,
    InferenceBackend,
    ModelType,
    TaskGroup,
)
from ..exceptions import (
    InvalidBenchmark,
    InvalidModel,
    NeedsEnvironmentVariable,
    NeedsExtraInstalled,
    NeedsSystemDependency,
)
from ..generation_utils import (
    apply_prompt,
    extract_few_shot_examples,
    raise_if_wrong_params,
)
from ..languages import get_all_languages
from ..logging_utils import get_pbar, log, log_once, no_terminal_output
from ..task_group_utils import (
    question_answering,
    sequence_classification,
    text_to_text,
    token_classification,
)
from ..tokenisation_utils import (
    apply_chat_template,
    get_bos_token,
    get_end_of_chat_token_ids,
    get_eos_token,
    get_first_label_token_mapping,
    get_pad_token,
    has_chat_template,
    should_prompts_be_stripped,
)
from ..types import ExtractLabelsFunction, Tokeniser
from ..utils import (
    clear_memory,
    create_model_cache_dir,
    get_hf_token,
    get_min_cuda_compute_capability,
    internet_connection_available,
    resolve_model_path,
    split_model_id,
)
from .hf import HuggingFaceEncoderModel, get_model_repo_info, load_hf_model_config

try:
    from transformers.tokenization_mistral_common import MistralCommonTokenizer
except ImportError:
    from transformers.tokenization_mistral_common import (
        MistralCommonBackend as MistralCommonTokenizer,
    )

if t.TYPE_CHECKING or importlib.util.find_spec("vllm") is not None:
    import vllm.config

    # MacOS/CPU installs an older version of vLLM, which doesn't have the attention
    # config
    if hasattr(vllm.config, "attention"):
        from vllm.config.attention import AttentionConfig

    from vllm import LLM, SamplingParams
    from vllm.distributed.parallel_state import (
        destroy_distributed_environment,
        destroy_model_parallel,
    )
    from vllm.lora.request import LoRARequest
    from vllm.sampling_params import StructuredOutputsParams

if t.TYPE_CHECKING or importlib.util.find_spec("ray") is not None:
    import ray


if t.TYPE_CHECKING:
    from datasets import DatasetDict
    from transformers.trainer import Trainer

    from ..data_models import BenchmarkConfig, DatasetConfig, Task


MODELS_REQUIRING_CUSTOM_ATTENTION_BACKENDS: dict[
    re.Pattern, t.Literal[*ATTENTION_BACKENDS]  # pyrefly: ignore[invalid-literal]
] = {
    re.compile(r".*gpt-oss.*", flags=re.IGNORECASE): "TRITON_ATTN",
    re.compile(r"google/gemma-3-1b.*", flags=re.IGNORECASE): "TRITON_ATTN",
    re.compile(r"google/gemma-3n.*", flags=re.IGNORECASE): "TRITON_ATTN",
    re.compile(r"google/gemma-3-(4|12|27)b.*", flags=re.IGNORECASE): "TRITON_ATTN",
    re.compile(r"PleIAs/Pleias-3b-Preview", flags=re.IGNORECASE): "TRITON_ATTN",
}


class VLLMModel(HuggingFaceEncoderModel):
    """A generative model using the vLLM inference framework."""

    fresh_model = False
    batching_preference = BatchingPreference.ALL_AT_ONCE
    high_priority = True
    allowed_params = {
        re.compile(r".*"): ["thinking", "no-thinking", "slow-tokenizer"],
        re.compile(r".*gpt-oss.*", flags=re.IGNORECASE): ["low", "medium", "high"],
    }

    def __init__(
        self,
        model_config: "ModelConfig",
        dataset_config: "DatasetConfig",
        benchmark_config: "BenchmarkConfig",
        log_metadata: bool = True,
    ) -> None:
        """Initialise the vLLM model.

        Args:
            model_config:
                The model configuration.
            dataset_config:
                The dataset configuration.
            benchmark_config:
                The benchmark configuration.
            log_metadata:
                Whether to log the model and dataset metadata.
        """
        if importlib.util.find_spec("vllm") is None:
            raise NeedsExtraInstalled(extra="generative")

        if torch.cuda.is_available() and shutil.which("nvcc") is None:
            raise NeedsSystemDependency(
                dependency="nvcc",
                instructions=(
                    "Please install the CUDA Toolkit from "
                    "https://developer.nvidia.com/cuda-downloads or ensure that NVCC "
                    "is available in your PATH."
                ),
            )

        if not torch.cuda.is_available() and (
            dataset_config.task.task_group
            in [
                TaskGroup.SEQUENCE_CLASSIFICATION,
                TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION,
            ]
            or dataset_config.task.uses_structured_output
        ):
            raise InvalidBenchmark(
                "We currently require CUDA to benchmark generative models on tasks "
                "that uses structured generation, which includes the current task "
                f"{dataset_config.task.name}. This is due to an xgrammar issue, which "
                "will hopefully be fixed soon."
            )

        raise_if_wrong_params(
            model_config=model_config, allowed_params=self.allowed_params
        )

        # Determine the attention backend to use:
        # Override for models that require a specific backend, otherwise use user's
        # choice from CLI (defaults to FLASHINFER)
        if hasattr(vllm.config, "attention"):
            for pattern, backend in MODELS_REQUIRING_CUSTOM_ATTENTION_BACKENDS.items():
                if re.search(pattern=pattern, string=model_config.model_id):
                    attention_backend = backend
                    break
            else:
                attention_backend = benchmark_config.attention_backend
        else:
            attention_backend = benchmark_config.attention_backend

        with no_terminal_output(disable=benchmark_config.verbose):
            model, tokeniser = load_model_and_tokeniser(
                model_config=model_config,
                benchmark_config=benchmark_config,
                attention_backend=attention_backend,
            )
        self._model: "LLM" = model
        self._tokeniser: Tokeniser = tokeniser

        # We specify `HuggingFaceEncoderModel` here instead of `VLLMModel`, as we want
        # to call the `__init__` method of the `BenchmarkModule` class.
        super(HuggingFaceEncoderModel, self).__init__(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
            log_metadata=log_metadata,
        )

        self.end_of_reasoning_token = get_end_of_reasoning_token(
            model=self._model, tokeniser=self._tokeniser, model_config=model_config
        )
        self.end_of_chat_token_ids = get_end_of_chat_token_ids(
            tokeniser=self._tokeniser, generative_type=self.generative_type
        )
        self.custom_stop_tokens = get_custom_stop_tokens(
            model=self._model,
            tokeniser=self._tokeniser,
            model_id=model_config.model_id,
            generative_type=self.generative_type,
        )

        self.buffer |= dict(
            first_label_token_mapping=get_first_label_token_mapping(
                dataset_config=self.dataset_config,
                model_config=self.model_config,
                tokeniser=self._tokeniser,
                generative_type=self.generative_type,
                log_metadata=self.log_metadata,
            )
        )
        if self.model_config.adapter_base_model_id is not None:
            if Path(self.model_config.model_id).exists():
                adapter_path = self.model_config.model_id
            else:
                adapter_path = snapshot_download(
                    repo_id=self.model_config.model_id,
                    revision=self.model_config.revision,
                    cache_dir=Path(self.model_config.model_cache_dir),
                )
            self.buffer["lora_request"] = LoRARequest(
                lora_name="adapter", lora_int_id=1, lora_path=adapter_path
            )

    def __del__(self) -> None:
        """Clean up the model and tokeniser."""
        try:
            if importlib.util.find_spec("vllm") is not None:
                clear_vllm()
        except ImportError:
            pass
        if hasattr(self, "_model"):
            del self._model
        if hasattr(self, "_tokeniser"):
            del self._tokeniser

    @property
    def generative_type(self) -> GenerativeType | None:
        """Get the generative type of the model.

        Returns:
            The generative type of the model, or None if it has not been set yet.
        """
        if not hasattr(self, "_tokeniser"):
            log_once(
                "The generative type of the model has not been set yet as the "
                "tokeniser has not been loaded.",
                level=logging.DEBUG,
            )
            return None
        elif self.benchmark_config.generative_type is not None:
            type_ = self.benchmark_config.generative_type
        elif self.model_config.param in {"thinking"}:
            type_ = GenerativeType.REASONING
        elif self.model_config.param in {"no-thinking"}:
            type_ = GenerativeType.INSTRUCTION_TUNED
        elif (
            hasattr(self, "end_of_reasoning_token")
            and self.end_of_reasoning_token is not None
        ):
            type_ = GenerativeType.REASONING
        elif (
            has_chat_template(tokeniser=self._tokeniser)
            or "instruct" in self.model_config.model_id.lower()
        ):
            type_ = GenerativeType.INSTRUCTION_TUNED
        else:
            type_ = GenerativeType.BASE
        log_once(
            f"Detected generative type {type_.name!r} for model "
            f"{self.model_config.model_id!r}",
            level=logging.DEBUG,
        )
        return type_

    @property
    def extract_labels_from_generation(self) -> ExtractLabelsFunction:
        """The function used to extract the labels from the generated output.

        Returns:
            The function used to extract the labels from the generated output.
        """
        match self.dataset_config.task.task_group:
            case (
                TaskGroup.SEQUENCE_CLASSIFICATION
                | TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION
            ):
                return partial(
                    sequence_classification.extract_labels_from_generation,
                    dataset_config=self.dataset_config,
                    model_config=self.model_config,
                    first_label_token_mapping=self.buffer["first_label_token_mapping"],
                )
            case TaskGroup.TEXT_TO_TEXT:
                return text_to_text.extract_labels_from_generation
            case TaskGroup.TOKEN_CLASSIFICATION:
                return partial(
                    token_classification.extract_labels_from_generation,
                    dataset_config=self.dataset_config,
                )
            case TaskGroup.QUESTION_ANSWERING:
                return question_answering.extract_labels_from_generation
            case _:
                raise NotImplementedError(
                    f"Unsupported task group: {self.dataset_config.task.task_group}."
                )

    def prepare_dataset(
        self, dataset: "DatasetDict", task: "Task", itr_idx: int
    ) -> "DatasetDict":
        """Prepare the dataset for the model.

        This includes things like tokenisation.

        Args:
            dataset:
                The dataset to prepare.
            task:
                The task to prepare the dataset for.
            itr_idx:
                The index of the dataset in the iterator.

        Returns:
            The prepared dataset.
        """
        if task.task_group == TaskGroup.QUESTION_ANSWERING:
            dataset = dataset.map(
                lambda examples: dict(
                    label=[
                        dict(
                            id=id,
                            answers=dict(
                                answer_start=answer_dct["answer_start"],
                                text=[
                                    answer_text.lower()
                                    for answer_text in answer_dct["text"]
                                ],
                            ),
                        )
                        for id, answer_dct in zip(examples["id"], examples["answers"])
                    ]
                ),
                batched=True,
                load_from_cache_file=False,
                keep_in_memory=True,
            )

        if self.benchmark_config.few_shot:
            few_shot_examples = extract_few_shot_examples(
                dataset=dataset,
                dataset_config=self.dataset_config,
                benchmark_config=self.benchmark_config,
                itr_idx=itr_idx,
            )
        else:
            few_shot_examples = list()

        dataset["test"] = dataset["test"].map(  # type: ignore[unsupported-operation]
            partial(
                apply_prompt,
                few_shot_examples=few_shot_examples,
                model_config=self.model_config,
                dataset_config=self.dataset_config,
                generative_type=self.generative_type,
                always_populate_text_field=True,
                tokeniser=self._tokeniser,
            ),
            batched=True,
            load_from_cache_file=False,
            keep_in_memory=True,
        )

        return dataset

    def generate(self, inputs: dict) -> "GenerativeModelOutput":
        """Generate outputs from the model.

        Args:
            inputs:
                A batch of inputs to pass through the model.

        Returns:
            The generated model outputs.

        Raises:
            InvalidBenchmark:
                If the dataset requires logprobs, but we could not get the first token
                of each label in the dataset.
        """
        # Get stopping tokens
        stop_tokens: list[str] = self.custom_stop_tokens.copy()
        if self.generative_type == GenerativeType.BASE:
            stop_tokens.append("\n\n")
        if self._tokeniser.pad_token_id is not None:
            assert isinstance(self._tokeniser.pad_token, str), (
                f"The pad token for the model {self.model_config.model_id!r} "
                f"is not a string, which is unexpected: {self._tokeniser.pad_token!r}."
            )
            stop_tokens.append(self._tokeniser.pad_token)
        if self._tokeniser.eos_token_id is not None:
            assert isinstance(self._tokeniser.eos_token, str), (
                f"The EOS token for the model {self.model_config.model_id!r} "
                f"is not a string, which is unexpected: {self._tokeniser.eos_token!r}."
            )
            stop_tokens.append(self._tokeniser.eos_token)
            if self._tokeniser.pad_token_id is None:
                self._tokeniser.pad_token_id = self._tokeniser.eos_token_id
                self._tokeniser.pad_token = self._tokeniser.eos_token
        if self.end_of_chat_token_ids is not None:
            end_of_chat_token = self._tokeniser.decode(
                list(self.end_of_chat_token_ids)
            ).strip()
            if end_of_chat_token:
                stop_tokens.append(end_of_chat_token)

        # Get the mapping from labels to the first token in the label. We call this each
        # time we generate a new dataset since the dataset config can change
        self.buffer["first_label_token_mapping"] = get_first_label_token_mapping(
            dataset_config=self.dataset_config,
            model_config=self.model_config,
            tokeniser=self._tokeniser,
            generative_type=self.generative_type,
            log_metadata=self.log_metadata,
        )
        if (
            not self.buffer["first_label_token_mapping"]
            and self.dataset_config.task.requires_logprobs
        ):
            raise InvalidBenchmark(
                "The dataset requires logprobs, but we encountered an error when "
                "trying to get the first token of each label in the dataset. You can "
                "try running this benchmark with the --verbose flag to see what the "
                "error was. Skipping this evaluation."
            )

        structured_generation_schema = None
        if (
            self.dataset_config.task.uses_structured_output
            or (self.dataset_config.task.uses_logprobs and self.dataset_config.labels)
        ) and self.generative_type == GenerativeType.REASONING:
            structured_outputs = None
            log_once(
                "The dataset uses structured output, but we are not using it as the "
                f"model {self.model_config.model_id!r} is a reasoning model.",
                level=logging.DEBUG,
            )
        elif self.dataset_config.task.uses_structured_output:
            ner_tag_names = list(self.dataset_config.prompt_label_mapping.values())
            keys_and_their_types: dict[str, t.Any] = {
                tag_name: (conlist(str, max_length=5), ...)
                for tag_name in ner_tag_names
            }
            answer_format_class = create_model("AnswerFormat", **keys_and_their_types)
            structured_generation_schema = answer_format_class.model_json_schema()
            log_once(
                "Using structured generation with the JSON schema: "
                f"{json.dumps(structured_generation_schema)}",
                level=logging.DEBUG,
            )
            structured_outputs = StructuredOutputsParams(
                json=structured_generation_schema
            )
        elif (
            self.dataset_config.task.uses_logprobs
            and self.dataset_config.labels
            and self.buffer.get("first_label_token_mapping", False)
        ):
            choice_labels = [
                self.dataset_config.prompt_label_mapping[label]
                for label in self.dataset_config.labels
            ]
            if isinstance(self.buffer["first_label_token_mapping"], dict):
                choice_labels = [
                    self.buffer["first_label_token_mapping"][label]
                    for label in choice_labels
                ]
            structured_outputs = StructuredOutputsParams(choice=choice_labels)
            log_once(
                "Using structured generation with the choices: "
                f"{structured_outputs.choice!r}.",
                level=logging.DEBUG,
            )
        else:
            structured_outputs = None
            log_once(
                "Not using structured generation as the dataset does not require it.",
                level=logging.DEBUG,
            )

        # Define the parameters used for vLLM generation
        generation_kwargs = GENERATION_KWARGS.copy()
        if (generation_config := self.model_config.generation_config) is not None:
            changed_params = generation_config.to_diff_dict()
            if "temperature" in changed_params:
                temperature = changed_params["temperature"]
                generation_kwargs["temperature"] = temperature
                log_once(
                    f"Using temperature={temperature} with the model "
                    f"{self.model_config.model_id!r} as specified in its "
                    "generation configuration.",
                    level=logging.DEBUG,
                )
            if "top_p" in changed_params:
                top_p = changed_params["top_p"]
                generation_kwargs["top_p"] = top_p
                log_once(
                    f"Using top_p={top_p} with the model "
                    f"{self.model_config.model_id!r} as specified in its "
                    "generation configuration.",
                    level=logging.DEBUG,
                )
            if "top_k" in changed_params:
                top_k = changed_params["top_k"]
                generation_kwargs["top_k"] = top_k
                log_once(
                    f"Using top_k={top_k} with the model "
                    f"{self.model_config.model_id!r} as specified in its "
                    "generation configuration.",
                    level=logging.DEBUG,
                )
            if "repetition_penalty" in changed_params:
                repetition_penalty = changed_params["repetition_penalty"]
                generation_kwargs["repetition_penalty"] = repetition_penalty
                log_once(
                    f"Using repetition_penalty={repetition_penalty} with the model "
                    f"{self.model_config.model_id!r} as specified in its "
                    "generation configuration.",
                    level=logging.DEBUG,
                )

        max_tokens: int = (
            REASONING_MAX_TOKENS
            if self.generative_type == GenerativeType.REASONING
            else self.dataset_config.max_generated_tokens
        )
        sampling_params = SamplingParams(
            max_tokens=max_tokens,
            logprobs=MAX_VLLM_LOGPROBS
            if self.buffer["first_label_token_mapping"]
            else None,
            temperature=generation_kwargs["temperature"],
            top_p=generation_kwargs["top_p"],
            top_k=int(generation_kwargs["top_k"]),
            repetition_penalty=generation_kwargs["repetition_penalty"],
            stop=[stop_token for stop_token in stop_tokens if stop_token],
            structured_outputs=structured_outputs,
        )

        # If any of the prompts are empty then we need to replace them with a BOS token
        # so that the vLLM model can generate from them
        prompts: c.Sequence[str] = inputs["text"]
        if any(len(prompt.strip()) == 0 for prompt in prompts):
            log("Found empty prompts, replacing with BOS token.", level=logging.DEBUG)
            prompts = [
                prompt
                if len(prompt.strip()) > 0
                else str(self._tokeniser.bos_token or "x")
                for prompt in prompts
            ]

        # Strip the prompts if the model's tokeniser requires it
        labels_to_be_generated = list(self.dataset_config.prompt_label_mapping.values())
        if len(labels_to_be_generated) == 0:
            labels_to_be_generated = ["negative", "positive"]
        if self.generative_type == GenerativeType.BASE and should_prompts_be_stripped(
            labels_to_be_generated=labels_to_be_generated, tokeniser=self._tokeniser
        ):
            log_once(
                f"Stripping prompts for model {self.model_config.model_id!r}.",
                level=logging.DEBUG,
            )
            prompts = [prompt.strip() for prompt in prompts]

        # Truncate the prompts if needed
        max_tokens_per_prompt = min(
            self._tokeniser.model_max_length, MAX_CONTEXT_LENGTH
        )
        max_tokens_per_prompt -= min(
            self.dataset_config.max_generated_tokens, max_tokens_per_prompt - 1
        )
        tokenized_prompts = self._tokeniser(
            text=prompts, max_length=max_tokens_per_prompt
        )
        if any(
            len(input_ids) >= max_tokens_per_prompt
            for input_ids in tokenized_prompts.input_ids
        ):
            log(
                f"Truncating prompts for the model {self.model_config.model_id!r} "
                f"to a maximum of {max_tokens_per_prompt:,} tokens.",
                level=logging.DEBUG,
            )
            match self.generative_type:
                case GenerativeType.BASE:
                    truncated_tokenized_prompts = self._tokeniser(
                        text=prompts, max_length=max_tokens_per_prompt, truncation=True
                    )
                    prompts = self._tokeniser.batch_decode(
                        sequences=truncated_tokenized_prompts.input_ids,
                        skip_special_tokens=True,
                    )
                case GenerativeType.INSTRUCTION_TUNED | GenerativeType.REASONING:
                    assert self.end_of_chat_token_ids is not None, (
                        "The end-of-chat token IDs should be set for instruction-tuned "
                        "and reasoning models."
                    )
                    end_of_chat_token = self._tokeniser.decode(
                        list(self.end_of_chat_token_ids)
                    )
                    prompt_segments: list[list[str]] = [
                        prompt.replace(self._tokeniser.bos_token, "").split(
                            end_of_chat_token
                        )
                        for prompt in prompts
                    ]
                    for num_few_shots_to_remove in range(
                        1, self.dataset_config.num_few_shot_examples + 1
                    ):
                        new_prompts = [
                            end_of_chat_token.join(
                                prompt_segment[2 * num_few_shots_to_remove :]
                            )
                            for prompt_segment in prompt_segments
                        ]
                        tokenized_prompts = self._tokeniser(
                            text=new_prompts, max_length=max_tokens_per_prompt
                        )
                        if all(
                            len(input_ids) < max_tokens_per_prompt
                            for input_ids in tokenized_prompts.input_ids
                        ):
                            prompts = new_prompts
                            break
                    else:
                        raise InvalidBenchmark(
                            "Truncation of prompts failed, some prompts are still too "
                            "long."
                        )
                case _:
                    raise InvalidBenchmark("The model type is not set!")
        else:
            log(
                f"Truncation of prompts for model {self.model_config.model_id!r} is "
                "not needed, so skipping truncation.",
                level=logging.DEBUG,
            )

        # Generate sequences using vLLM
        input_is_a_test = len(prompts) == 1 and len(set(prompts[0])) == 1
        num_attempts = 3
        truncation_attempts = 1
        for _ in range(num_attempts):
            try:
                raw_outputs = self._model.generate(
                    prompts=prompts,
                    sampling_params=sampling_params,
                    use_tqdm=False if input_is_a_test else get_pbar,
                    lora_request=self.buffer.get("lora_request"),
                )
                break
            except TypeError as e:
                log(
                    f"Encountered error during vLLM generation: {str(e)}. Retrying...",
                    level=logging.DEBUG,
                )
                sleep(1)
            except (ValueError, RuntimeError) as e:
                # Truncate the prompts if they are too long for the model
                truncate_error_messages = [
                    r"prompt \(length [0-9]+\) is longer than the maximum model length",
                    "Sampled token IDs exceed the max model length",
                ]
                if any(
                    re.search(pattern, str(e), flags=re.IGNORECASE) is not None
                    for pattern in truncate_error_messages
                ):
                    log(
                        "Prompts are too long, so truncating them and trying again...",
                        level=logging.WARNING,
                    )
                    log(f"The error message was: {str(e)}", level=logging.DEBUG)

                    # If we have already tried truncating the prompts a few times, then
                    # we truncate a bit more aggressively
                    extra_truncation = 50 * truncation_attempts
                    truncation_attempts += 1

                    tokenized_prompts = self._tokeniser(
                        text=prompts,
                        truncation=True,
                        max_length=max(
                            min(self._tokeniser.model_max_length, MAX_CONTEXT_LENGTH)
                            - max_tokens
                            - extra_truncation,
                            0,
                        ),
                    )
                    prompts = self._tokeniser.batch_decode(
                        sequences=tokenized_prompts.input_ids, skip_special_tokens=True
                    )
                else:
                    raise InvalidBenchmark(
                        f"An error occurred during vLLM generation: {str(e)}"
                    ) from e
        else:
            raise InvalidBenchmark(
                f"Could not generate sequences after {num_attempts} attempts."
            )

        # When we shorten the prompts then some residual model outputs persist, so we
        # need to filter these out
        num_extra_outputs = len(raw_outputs) - len(prompts)
        if num_extra_outputs > 0:
            raw_outputs = raw_outputs[num_extra_outputs:]
            if not all(
                raw_output.prompt == prompt
                for raw_output, prompt in zip(raw_outputs, prompts)
            ):
                raise InvalidBenchmark(
                    f"The prompts and the model outputs do not match. There were "
                    f"{num_extra_outputs!r} extra outputs."
                )
            else:
                log(
                    f"Filtered out {num_extra_outputs:,} extra outputs from the model, "
                    "which occured as we interupted the generation when we truncated "
                    "the prompts.",
                    level=logging.DEBUG,
                )

        # Parse the raw model outputs. We keep the special tokens for now, as we need
        # them to potentially remove reasoning content and stop tokens
        completion_ids: c.Sequence[c.Sequence[int]] = [
            list(output.outputs[0].token_ids) for output in raw_outputs
        ]
        completions = self._tokeniser.batch_decode(
            sequences=[list(completion_id) for completion_id in completion_ids],
            skip_special_tokens=False,
        )
        if (
            self.end_of_reasoning_token is not None
            and self.generative_type == GenerativeType.REASONING
        ):
            num_samples_without_eor_token = 0
            for idx in range(len(completions)):
                if (
                    isinstance(self.end_of_reasoning_token, str)
                    and self.end_of_reasoning_token in completions[idx]
                ):
                    completions[idx] = completions[idx].split(
                        self.end_of_reasoning_token
                    )[-1]
                elif isinstance(
                    self.end_of_reasoning_token, re.Pattern
                ) and self.end_of_reasoning_token.search(completions[idx]):
                    completions[idx] = self.end_of_reasoning_token.split(
                        completions[idx]
                    )[-1]
                else:
                    num_samples_without_eor_token += 1
                    completions[idx] = ""
            if num_samples_without_eor_token > 0:
                log_once(
                    f"The model {self.model_config.model_id!r} is a reasoning "
                    "model, but the generated output did not contain the end of "
                    f"reasoning token ({self.end_of_reasoning_token!r}) in "
                    f"{num_samples_without_eor_token:,}/{len(completions):,} of "
                    "the samples. Using an empty string for all these samples "
                    "instead.",
                    level=(
                        logging.WARNING
                        if num_samples_without_eor_token / len(completions) > 0.5
                        else logging.DEBUG
                    ),
                )
        stop_token_pattern = re.compile(
            "|".join(re.escape(stop_token) for stop_token in stop_tokens)
        )
        completions = [
            re.split(pattern=stop_token_pattern, string=completion)[0].strip()
            for completion in completions
        ]

        # Remove all the special tokens from the completions, if any are present
        completion_ids = self._tokeniser(text=completions).input_ids
        completions = self._tokeniser.batch_decode(
            sequences=completion_ids, skip_special_tokens=True
        )

        # Sanity check
        if len(completions) != len(prompts):
            raise InvalidBenchmark(
                f"Expected {len(prompts):,} completions, but got {len(completions):,}."
            )

        # Add logprobs scores to the output
        if self.buffer["first_label_token_mapping"]:
            scores: c.Sequence[c.Sequence[c.Sequence[tuple[str, float]]]] = [
                [
                    [
                        (obj.decoded_token or "", obj.logprob)
                        for obj in token_logprobs_dict.values()
                    ]
                    for token_logprobs_dict in raw_output.outputs[0].logprobs or list()
                ]
                for raw_output in raw_outputs
            ]
            output = GenerativeModelOutput(sequences=completions, scores=scores)
        else:
            output = GenerativeModelOutput(sequences=completions)

        return output

    @classmethod
    def model_exists(
        cls, model_id: str, benchmark_config: "BenchmarkConfig"
    ) -> bool | NeedsExtraInstalled | NeedsEnvironmentVariable:
        """Check if a model exists.

        Args:
            model_id:
                The model ID.
            benchmark_config:
                The benchmark configuration.

        Returns:
            Whether the model exists, or an error describing why we cannot check
            whether the model exists.
        """
        using_api = (
            benchmark_config.api_base is not None
            or benchmark_config.api_version is not None
        )
        if using_api:
            return False

        model_id_components = split_model_id(model_id=model_id)
        model_id = model_id_components.model_id
        revision = model_id_components.revision

        model_info = get_model_repo_info(
            model_id=model_id,
            revision=revision,
            api_key=benchmark_config.api_key,
            cache_dir=benchmark_config.cache_dir,
            trust_remote_code=benchmark_config.trust_remote_code,
            requires_safetensors=benchmark_config.requires_safetensors,
            run_with_cli=benchmark_config.run_with_cli,
        )
        return (
            model_info is not None
            and model_info.pipeline_tag in GENERATIVE_PIPELINE_TAGS
        )

    @classmethod
    def get_model_config(
        cls, model_id: str, benchmark_config: "BenchmarkConfig"
    ) -> "ModelConfig":
        """Fetch the model configuration.

        Args:
            model_id:
                The model ID.
            benchmark_config:
                The benchmark configuration.

        Returns:
            The model configuration.
        """
        model_id_components = split_model_id(model_id=model_id)
        model_info = get_model_repo_info(
            model_id=model_id_components.model_id,
            revision=model_id_components.revision,
            api_key=benchmark_config.api_key,
            cache_dir=benchmark_config.cache_dir,
            trust_remote_code=benchmark_config.trust_remote_code,
            requires_safetensors=benchmark_config.requires_safetensors,
            run_with_cli=benchmark_config.run_with_cli,
        )
        if model_info is None:
            raise InvalidModel(f"The model {model_id!r} could not be found.")

        try:
            generation_config = GenerationConfig.from_pretrained(
                pretrained_model_name=model_id_components.model_id,
                revision=model_id_components.revision,
                cache_dir=benchmark_config.cache_dir,
                token=benchmark_config.api_key,
            )
        except OSError:
            generation_config = None

        language_mapping = get_all_languages()
        language_codes = list(language_mapping.keys())

        model_config = ModelConfig(
            model_id=model_id_components.model_id,
            revision=model_id_components.revision,
            param=model_id_components.param,
            task=model_info.pipeline_tag,
            languages=[
                language_mapping[tag]
                for tag in model_info.tags
                if tag in language_codes
            ],
            merge=any(tag in model_info.tags for tag in MERGE_TAGS),
            inference_backend=InferenceBackend.VLLM,
            model_type=ModelType.GENERATIVE,
            fresh=False,
            model_cache_dir=create_model_cache_dir(
                cache_dir=benchmark_config.cache_dir, model_id=model_id
            ),
            adapter_base_model_id=model_info.adapter_base_model_id,
            generation_config=generation_config,
        )

        return model_config

    @property
    def data_collator(self) -> c.Callable[[list[dict[str, t.Any]]], dict[str, t.Any]]:
        """The data collator used to prepare samples during finetuning.

        Returns:
            The data collator.
        """
        raise NotImplementedError(
            "The `data_collator` property has not been implemented for vLLM models."
        )

    @property
    def trainer_class(self) -> t.Type["Trainer"]:
        """The Trainer class to use for finetuning.

        Returns:
            The Trainer class.
        """
        raise NotImplementedError(
            "The `trainer_class` property has not been implemented for vLLM models."
        )


def load_model_and_tokeniser(
    model_config: "ModelConfig",
    benchmark_config: "BenchmarkConfig",
    attention_backend: t.Literal[
        *ATTENTION_BACKENDS  # pyrefly: ignore[invalid-literal]
    ],
) -> tuple["LLM", Tokeniser]:
    """Load the model and tokeniser.

    Args:
        model_config:
            The model configuration.
        benchmark_config:
            The benchmark configuration.
        attention_backend:
            The attention backend to use.

    Returns:
        A pair (model, tokeniser), with the loaded model and tokeniser
    """
    # Prefer base model ID if the model is an adapter - the adapter will be added on
    # during inference in this case
    model_id = model_config.adapter_base_model_id or model_config.model_id
    revision = (
        model_config.revision if model_config.adapter_base_model_id is None else "main"
    )

    hf_model_config = load_hf_model_config(
        model_id=model_id,
        num_labels=0,
        id2label=HashableDict(),
        label2id=HashableDict(),
        revision=revision,
        model_cache_dir=model_config.model_cache_dir,
        api_key=benchmark_config.api_key,
        trust_remote_code=benchmark_config.trust_remote_code,
        run_with_cli=benchmark_config.run_with_cli,
    )

    # Start with dtype being the "auto" vLLM dtype
    dtype: str | torch.dtype = "auto"

    # Choose bf16 over fp16 if the model is a fp32 model and the GPU supports it
    if hf_model_config.dtype == torch.float32:
        if torch.cuda.is_bf16_supported():
            log(
                "You are loading a model with dtype FP32, which we will convert to "
                "BF16 as FP32 is not supported by vLLM and BF16 is supported by your "
                "GPU.",
                level=logging.WARNING,
            )
            dtype = torch.bfloat16
        else:
            log(
                "You are loading a model with dtype FP32, which we will convert to "
                "FP16 as FP32 is not supported by vLLM and BF16 is not supported by "
                "your GPU.",
                level=logging.WARNING,
            )
            dtype = torch.float16

    # If the model is a bf16 model, we need to check the CUDA compute capability
    if hf_model_config.dtype == torch.bfloat16:
        min_cuda_compute_capability = get_min_cuda_compute_capability()
        required_capability = VLLM_BF16_MIN_CUDA_COMPUTE_CAPABILITY

        if min_cuda_compute_capability is not None:
            if min_cuda_compute_capability < required_capability:
                log(
                    f"You are loading a model with dtype {hf_model_config.dtype}, "
                    "which vLLM only supports for CUDA devices with CUDA compute "
                    f"capability >={required_capability}. You are using one or more "
                    f"devices with compute capability {min_cuda_compute_capability}. "
                    "Setting dtype to float16 instead.",
                    level=logging.WARNING,
                )
                dtype = torch.float16

    quantization = None
    if hasattr(hf_model_config, "quantization_config"):
        quantization = hf_model_config.quantization_config.get("quant_method")

    # The quantised models require extra dependencies
    if quantization == "gptq" and (
        importlib.util.find_spec("auto_gptq") is None
        or importlib.util.find_spec("optimum") is None
    ):
        raise NeedsExtraInstalled(extra="quantization")
    if quantization == "awq" and importlib.util.find_spec("awq") is None:
        raise NeedsExtraInstalled(extra="quantization")

    # If the model is a quantized model, let vLLM decide the dtype
    if quantization is not None:
        log(
            f"You are loading a quantized model with quantization {quantization}. "
            "Forcing the vLLM dtype to 'auto'",
            level=logging.WARNING,
        )
        dtype = "auto"

    if model_config.adapter_base_model_id is not None:
        download_dir = str(Path(model_config.model_cache_dir) / "base_model")
    else:
        download_dir = str(model_config.model_cache_dir)

    potential_max_model_length_config_names = [
        "max_position_embeddings",
        "max_sequence_length",
        "model_max_length",
        "n_positions",
    ]
    true_max_model_len_candidates: list[int] = list()
    for config_name in potential_max_model_length_config_names:
        if hasattr(hf_model_config, config_name):
            model_len = getattr(hf_model_config, config_name)
            if model_len is not None:
                true_max_model_len_candidates.append(model_len)

    if len(true_max_model_len_candidates) > 0:
        true_max_model_len = min(true_max_model_len_candidates)
    else:
        true_max_model_len = MAX_CONTEXT_LENGTH

    tokeniser = load_tokeniser(
        model_id=model_config.model_id,
        revision=model_config.revision,
        adapter_base_model_id=model_config.adapter_base_model_id,
        trust_remote_code=benchmark_config.trust_remote_code,
        model_max_length=true_max_model_len,
        model_config=model_config,
        token=get_hf_token(api_key=benchmark_config.api_key),
    )
    vllm_params = get_vllm_tokenisation_params(
        tokeniser=tokeniser, model_config=model_config
    )

    # MacOS/CPU installs an older version of vLLM, which doesn't have the attention
    # config
    if hasattr(vllm.config, "attention"):
        vllm_params["attention_config"] = AttentionConfig(backend=attention_backend)

    clear_vllm()

    distributed_executor_backend, tensor_parallel_size, pipeline_parallel_size = (
        select_backend_and_parallelism()
    )

    try:
        model_location = (
            model_id
            if internet_connection_available() or Path(model_id).is_dir()
            else resolve_model_path(download_dir=download_dir)
        )

        max_model_len = min(
            true_max_model_len, MAX_CONTEXT_LENGTH + REASONING_MAX_TOKENS
        )
        model = LLM(
            model=model_location,
            tokenizer=model_location,
            gpu_memory_utilization=benchmark_config.gpu_memory_utilization,
            max_model_len=max_model_len,
            max_num_batched_tokens=max_model_len,
            download_dir=download_dir,
            trust_remote_code=benchmark_config.trust_remote_code,
            revision=revision,
            seed=4242,
            distributed_executor_backend=distributed_executor_backend,
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            disable_custom_all_reduce=True,
            quantization=quantization,
            dtype=dtype,
            enforce_eager=True,
            # TEMP: Prefix caching isn't supported with sliding window in vLLM yet,
            # so we disable it for now
            enable_prefix_caching=False,
            enable_lora=model_config.adapter_base_model_id is not None,
            max_lora_rank=256,
            **vllm_params,
        )
    except (RuntimeError, ValueError, OSError) as e:
        if "awaiting a review from the repo authors" in str(e):
            raise InvalidModel(
                f"The model {model_id!r} is awaiting a review from the repository "
                "authors. Please try again later."
            ) from e
        elif "trust_remote_code" in str(e):
            raise InvalidModel(
                f"Loading the model {model_id!r} needs to trust remote code. "
                "If you trust the suppliers of this model, then you can enable "
                "this by setting the `--trust-remote-code` flag."
            ) from e
        elif "See stack trace for root cause." in str(
            e
        ) or "See root cause above." in str(e):
            msg = (
                f"The model {model_id!r} could not be loaded, but vLLM did not "
                "mention exactly what happened. "
            )
            msg += (
                (
                    "Since you're running in verbose mode, you might see a descriptive "
                    "error above already. Note however that if the error message urges "
                    "you to use the attention backend 'FLEX_ATTENTION', please try "
                    "setting it to 'TRITON_ATTN' instead using the "
                    "`--attention-backend` CLI argument, as that often solves the "
                    "issue, whereas 'FLEX_ATTENTION' usually doesn't. If you don't "
                    "see any descriptive error above, then you can try "
                )
                if benchmark_config.verbose
                else "Try "
            )
            msg += (
                "re-running the benchmark with the environment variable `FULL_LOG` "
                "set to `1` to see the full stack trace. E.g., "
                f"`FULL_LOG=1 euroeval --model {model_id}`."
            )
            raise InvalidModel(msg) from e
        raise InvalidModel(
            f"The model {model_id!r} could not be loaded. The error was {e!r}."
        ) from e

    model.config = hf_model_config

    return model, tokeniser


def load_tokeniser(
    model_id: str,
    revision: str,
    adapter_base_model_id: str | None,
    trust_remote_code: bool,
    model_max_length: int,
    model_config: "ModelConfig",
    token: str | bool,
) -> Tokeniser:
    """Load the tokeniser.

    Args:
        model_id:
            The model identifier.
        revision:
            The revision of the model.
        adapter_base_model_id:
            The base model ID for the adapter model. Can be None if the model is not an
            adapter model.
        trust_remote_code:
            Whether to trust remote code.
        model_max_length:
            The maximum length of the model.
        model_config:
            The model configuration.
        token:
            The Hugging Face API token.

    Returns:
        The loaded tokeniser.
    """
    revision = revision if adapter_base_model_id is None else "main"
    config = AutoConfig.from_pretrained(
        adapter_base_model_id or model_id,
        revision=revision,
        cache_dir=model_config.model_cache_dir,
        token=token,
        trust_remote_code=trust_remote_code,
        local_files_only=not internet_connection_available(),
    )
    num_retries = 5
    for _ in range(num_retries):
        try:
            # Mistral instruction-tuned models need a custom tokeniser
            if model_id.startswith("mistralai/") and "base" not in model_id.lower():
                tokeniser = MistralCommonTokenizer.from_pretrained(
                    model_id,
                    padding_side="left",
                    truncation_side="left",
                    model_max_length=model_max_length,
                    token=token,
                )
                break
            tokeniser = AutoTokenizer.from_pretrained(
                model_id,
                revision=revision,
                use_fast=False if model_config.param == "slow-tokenizer" else True,
                verbose=False,
                trust_remote_code=trust_remote_code,
                padding_side="left",
                truncation_side="left",
                model_max_length=model_max_length,
                cache_dir=model_config.model_cache_dir,
                config=config,
                token=token,
                local_files_only=not internet_connection_available(),
            )
            break
        except (json.JSONDecodeError, OSError, TypeError) as e:
            if adapter_base_model_id is None or model_id == adapter_base_model_id:
                raise InvalidModel(
                    f"Could not load tokeniser for model {model_id!r}. The error was "
                    f"{str(e)}."
                ) from e
            log(
                f"Could not load tokeniser for {model_id!r}. Falling back to "
                f"{adapter_base_model_id!r}.",
                level=logging.DEBUG,
            )
            model_id = adapter_base_model_id
        except (TimeoutError, RequestError):
            log(
                f"Couldn't load tokeniser for {model_id!r}. Retrying.",
                level=logging.WARNING,
            )
            sleep(5)
            continue
        except (KeyError, ValueError) as e:
            if "mistral" in str(e).lower():
                tokeniser = MistralCommonTokenizer.from_pretrained(
                    model_id,
                    padding_side="left",
                    truncation_side="left",
                    model_max_length=model_max_length,
                    token=token,
                )
                break
            raise InvalidModel(
                f"Could not load tokeniser for model {model_id!r}. The error was "
                f"{str(e)}."
            ) from e
    else:
        raise InvalidModel(
            f"Could not load tokeniser for model {model_id!r} after {num_retries} "
            "attempts."
        )

    # Ensure that BOS, EOS and PAD tokens are set
    if not isinstance(tokeniser, MistralCommonTokenizer):
        tokeniser.bos_token, tokeniser.bos_token_id = get_bos_token(tokeniser=tokeniser)
        tokeniser.eos_token, tokeniser.eos_token_id = get_eos_token(tokeniser=tokeniser)
        tokeniser.pad_token, tokeniser.pad_token_id = get_pad_token(tokeniser=tokeniser)

    return tokeniser


def clear_vllm() -> None:
    """Clear the GPU memory used by the vLLM model, enabling re-initialisation."""
    with contextlib.suppress(ValueError):
        destroy_model_parallel()
        destroy_distributed_environment()
    with contextlib.suppress(AssertionError):
        torch.distributed.destroy_process_group()
    clear_memory()


def get_end_of_reasoning_token(
    model: "LLM", tokeniser: Tokeniser, model_config: "ModelConfig"
) -> str | re.Pattern | None:
    """Get the end-of-reasoning token for a generative model.

    Args:
        model:
            The vLLM model.
        tokeniser:
            The tokeniser.
        model_config:
            The model configuration.

    Returns:
        The end of reasoning token, or None if it could not be found.
    """
    model_id = model_config.model_id

    # Create a prompt to check if the model uses the reasoning tokens
    prompt = "What is your name?"
    if has_chat_template(tokeniser=tokeniser):
        extra_kwargs = dict()
        if model_config.param in {"thinking", "no-thinking"}:
            extra_kwargs["enable_thinking"] = model_config.param == "thinking"
        templated_prompt = apply_chat_template(
            conversation=[dict(role="user", content=prompt)],
            tokeniser=tokeniser,
            tokenise=False,
            add_generation_prompt=True,
            **extra_kwargs,
        )
        assert isinstance(templated_prompt, str)
        prompt = templated_prompt

    # Check that the beginning-of-reasoning token is actually used by the model
    output = model.generate(
        prompts=[prompt], sampling_params=SamplingParams(max_tokens=10), use_tqdm=False
    )[0]
    completion = tokeniser.decode(token_ids=output.outputs[0].token_ids)
    bor_reasoning_matches = [
        (bor_token, eor_token)
        for bor_token, eor_token in REASONING_TOKENS
        if (
            (
                isinstance(bor_token, str)
                and (bor_token in prompt or bor_token in completion)
            )
            or (
                isinstance(bor_token, re.Pattern)
                and (
                    bor_token.search(prompt) is not None
                    or bor_token.search(completion) is not None
                )
            )
        )
    ]
    if not bor_reasoning_matches:
        log_once(
            f"The model {model_id!r} did not generate any beginning-of-reasoning "
            "tokens in the prompt or the completion. Assuming the model is not a "
            "reasoning model.",
            level=logging.DEBUG,
        )
        return None

    # Check that the end-of-reasoning token is actually used by the model
    output = model.generate(
        prompts=[prompt],
        sampling_params=SamplingParams(max_tokens=REASONING_MAX_TOKENS),
        use_tqdm=False,
    )[0]
    completion = tokeniser.decode(token_ids=output.outputs[0].token_ids)
    eor_reasoning_matches = [
        (bor_token, eor_token)
        for bor_token, eor_token in bor_reasoning_matches
        if (
            (isinstance(eor_token, str) and eor_token in completion)
            or (
                isinstance(eor_token, re.Pattern)
                and eor_token.search(completion) is not None
            )
        )
    ]
    if not eor_reasoning_matches:
        log_once(
            f"The model {model_id!r} did not generate any end-of-reasoning "
            "tokens in the prompt or the completion, even though it generated "
            "the beginning-of-reasoning tokens "
            f"{[bor_token for bor_token, _ in bor_reasoning_matches]!r}. "
            "This is probably not correct, so please report this issue.",
            level=logging.WARNING,
        )
        return None

    if len(eor_reasoning_matches) > 1:
        log_once(
            f"Found multiple reasoning tokens {eor_reasoning_matches} for "
            f"model {model_id!r}. Using {eor_reasoning_matches[0]!r} as "
            "the reasoning token. If this is not the correct reasoning token, "
            "please report this issue.",
            level=logging.WARNING,
        )

    bor_token, eor_token = eor_reasoning_matches[0]

    bor_token_logging: str = (
        bor_token if isinstance(bor_token, str) else bor_token.pattern
    )
    eor_token_logging: str = (
        eor_token if isinstance(eor_token, str) else eor_token.pattern
    )
    log_once(
        f"Detected beginning-of-reasoning token {bor_token_logging!r} and "
        f"end-of-reasoning token {eor_token_logging!r} for model {model_id!r}.",
        level=logging.DEBUG,
    )

    return eor_token


def get_custom_stop_tokens(
    model: "LLM",
    tokeniser: Tokeniser,
    model_id: str,
    generative_type: GenerativeType | None,
) -> list[str]:
    """Get the stop tokens for a generative model.

    Args:
        model:
            The vLLM model.
        tokeniser:
            The tokeniser.
        model_id:
            The model ID.
        generative_type:
            The generative type of the model.

    Returns:
        A list of stop tokens.
    """
    candidate_stop_tokens = CUSTOM_STOP_TOKENS

    prompt = "Hello"
    if has_chat_template(tokeniser=tokeniser):
        templated_prompt = apply_chat_template(
            conversation=[dict(role="user", content=prompt)],
            tokeniser=tokeniser,
            tokenise=False,
            add_generation_prompt=True,
            enable_thinking=generative_type == GenerativeType.REASONING,
        )
        assert isinstance(templated_prompt, str)
        prompt = templated_prompt

    max_tokens = (
        REASONING_MAX_TOKENS if generative_type == GenerativeType.REASONING else 10
    )
    output = model.generate(
        prompts=[prompt],
        sampling_params=SamplingParams(max_tokens=max_tokens, temperature=0.0),
        use_tqdm=False,
    )[0]
    completion = tokeniser.decode(token_ids=output.outputs[0].token_ids)

    stop_tokens = [
        stop_token
        for stop_token in candidate_stop_tokens
        if stop_token in prompt or stop_token in completion
    ]
    if stop_tokens:
        log(
            f"Found the following custom stop tokens for model {model_id!r}: "
            f"{stop_tokens}.",
            level=logging.DEBUG,
        )
    else:
        log(f"Found no custom stop tokens for model {model_id!r}.", level=logging.DEBUG)

    return stop_tokens


def get_vllm_tokenisation_params(
    tokeniser: Tokeniser, model_config: "ModelConfig"
) -> dict[str, t.Any]:
    """Get the tokenisation parameters for vLLM.

    Args:
        tokeniser:
            The tokeniser.
        model_config:
            The model configuration.

    Returns:
        A dictionary of tokenisation parameters to pass to vLLM.
    """
    if isinstance(tokeniser, MistralCommonTokenizer):
        tokeniser_mode = "mistral"
    elif model_config.param == "slow-tokenizer":
        tokeniser_mode = "slow"
    else:
        tokeniser_mode = "auto"

    if isinstance(tokeniser, MistralCommonTokenizer):
        config_format = "mistral"
    else:
        config_format = "auto"

    if isinstance(tokeniser, MistralCommonTokenizer):
        load_format = "mistral"
    else:
        load_format = "auto"

    return dict(
        tokenizer_mode=tokeniser_mode,
        config_format=config_format,
        load_format=load_format,
    )


def select_backend_and_parallelism() -> tuple[str, int, int]:
    """Determine the distributed backend and parallelism for vLLM.

    Returns:
        Tuple containing:
        - backend (str): "ray" if multi-node Ray is available, else "mp".
        - tensor_parallel_size (int): Number of GPUs per node.
        - pipeline_parallel_size (int): Number of stages across nodes.
    """
    if not torch.cuda.is_available():
        return "mp", 1, 1

    if not ray.is_initialized():
        try:
            ray.init(address="auto", ignore_reinit_error=True)
        except Exception as e:
            if "could not find any running ray instance" not in str(e).lower():
                log_once(
                    f"Ray initialisation failed with a {type(e)} exception: {e}",
                    level=logging.DEBUG,
                )

    is_ray = ray.is_initialized()
    local_gpu_count = torch.cuda.device_count()

    if is_ray:
        resources = ray.cluster_resources()
        total_gpus = int(resources.get("GPU", 0))
    else:
        total_gpus = local_gpu_count

    using_multiple_nodes = total_gpus > local_gpu_count
    if is_ray and using_multiple_nodes:
        distributed_executor_backend = "ray"
        tensor_parallel_size = local_gpu_count if local_gpu_count > 0 else 1
        pipeline_parallel_size = max(1, total_gpus // tensor_parallel_size)
        log_once(
            f"Detected a multi-node setup with {pipeline_parallel_size:,} nodes, each "
            f"with {tensor_parallel_size:,} GPUs, so using `ray` as the "
            "distributed backend.",
            level=logging.DEBUG,
        )
    else:
        distributed_executor_backend = "mp"
        tensor_parallel_size = local_gpu_count if local_gpu_count > 0 else 1
        pipeline_parallel_size = 1
        log_once(
            f"Detected a single-node setup with {tensor_parallel_size:,} GPUs, "
            "so using the multiprocessing distributed backend.",
            level=logging.DEBUG,
        )

    return distributed_executor_backend, tensor_parallel_size, pipeline_parallel_size



"""All dataset configurations used in EuroEval."""

from pathlib import Path

from ..data_models import DatasetConfig
from ..languages import get_all_languages
from ..tasks import SPEED
from ..utils import load_custom_datasets_module
from .albanian import *  # noqa: F403
from .bosnian import *  # noqa: F403
from .bulgarian import *  # noqa: F403
from .catalan import *  # noqa: F403
from .croatian import *  # noqa: F403
from .czech import *  # noqa: F403
from .danish import *  # noqa: F403
from .dutch import *  # noqa: F403
from .english import *  # noqa: F403
from .estonian import *  # noqa: F403
from .faroese import *  # noqa: F403
from .finnish import *  # noqa: F403
from .french import *  # noqa: F403
from .german import *  # noqa: F403
from .greek import *  # noqa: F403
from .hungarian import *  # noqa: F403
from .icelandic import *  # noqa: F403
from .italian import *  # noqa: F403
from .latvian import *  # noqa: F403
from .lithuanian import *  # noqa: F403
from .norwegian import *  # noqa: F403
from .polish import *  # noqa: F403
from .portuguese import *  # noqa: F403
from .romanian import *  # noqa: F403
from .serbian import *  # noqa: F403
from .slovak import *  # noqa: F403
from .slovene import *  # noqa: F403
from .spanish import *  # noqa: F403
from .swedish import *  # noqa: F403
from .ukrainian import *  # noqa: F403


def get_all_dataset_configs(custom_datasets_file: Path) -> dict[str, DatasetConfig]:
    """Get a mapping of all the dataset configurations.

    Args:
        custom_datasets_file:
            A path to a Python file containing custom dataset configurations.

    Returns:
        A mapping between names of datasets and their configurations.
    """
    globals_dict = globals()
    module = load_custom_datasets_module(custom_datasets_file=custom_datasets_file)
    if module is not None:
        globals_dict |= vars(module)
    dataset_configs = [
        cfg
        for cfg in globals_dict.values()
        if isinstance(cfg, DatasetConfig) and cfg.task != SPEED
    ]
    assert len(dataset_configs) == len({cfg.name for cfg in dataset_configs}), (
        "There are duplicate dataset configurations. Please ensure that each dataset "
        "has a unique name."
    )
    mapping = {cfg.name: cfg for cfg in dataset_configs}
    return mapping


def get_dataset_config(dataset_name: str, custom_datasets_file: Path) -> DatasetConfig:
    """Get the dataset configuration for a dataset.

    Args:
        dataset_name:
            The name of the dataset.
        custom_datasets_file:
            A path to a Python file containing custom dataset configurations.

    Returns:
        The dataset configuration.

    Raises:
        ValueError:
            If the dataset is not found.
    """
    dataset_configs = get_all_dataset_configs(custom_datasets_file=custom_datasets_file)
    if dataset_name not in dataset_configs:
        raise ValueError(f"No dataset config found for dataset {dataset_name}.")
    return dataset_configs[dataset_name]


SPEED_CONFIG = DatasetConfig(
    name="speed",
    pretty_name="",
    source="",
    task=SPEED,
    languages=list(get_all_languages().values()),
    _logging_string="the speed estimation benchmark",
)



"""Metrics based on a scikit-learn Pipeline."""

import collections.abc as c
import logging
import typing as t
from pathlib import Path

import cloudpickle
import huggingface_hub as hf_hub
import numpy as np
from scipy.special import expit as sigmoid

from ..exceptions import InvalidBenchmark
from ..logging_utils import log, no_terminal_output
from ..utils import unscramble
from .base import Metric

if t.TYPE_CHECKING:
    from datasets.arrow_dataset import Dataset
    from sklearn.pipeline import Pipeline

    from ..data_models import BenchmarkConfig, DatasetConfig


T = t.TypeVar("T", bound=int | float | str | bool)


class PreprocessingFunction(t.Protocol):
    """A protocol for a preprocessing function."""

    def __call__(
        self, predictions: c.Sequence[int], dataset: "Dataset"
    ) -> c.Sequence[int]:
        """Preprocess the model predictions before they are passed to the pipeline.

        Args:
            predictions:
                The model predictions.
            dataset:
                The dataset used for evaluation. This is only used in case any
                additional metadata is used to compute the metrics.

        Returns:
            The preprocessed model predictions.
        """
        ...


class PipelineMetric(Metric):
    """Load a scikit-learn pipeline and use it to get scores from the predictions."""

    def __init__(
        self,
        name: str,
        pretty_name: str,
        pipeline_repo: str,
        pipeline_scoring_function: c.Callable[["Pipeline", c.Sequence], float],
        pipeline_file_name: str = "pipeline.pkl",
        preprocessing_fn: PreprocessingFunction | None = None,
        postprocessing_fn: c.Callable[[float], tuple[float, str]] | None = None,
    ) -> None:
        """Initialise the pipeline transform metric.

        Args:
            name:
                The name of the metric in snake_case.
            pretty_name:
                The pretty name of the metric, used for display purposes.
            pipeline_repo:
                The Hugging Face repository ID of the scikit-learn pipeline to load.
            pipeline_scoring_method:
                The method to use for scoring the predictions with the pipeline. Takes
                a 1D sequence of predictions and returns a float score.
            pipeline_file_name (optional):
                The name of the file to download from the Hugging Face repository.
                Defaults to "pipeline.joblib".
            preprocessing_fn (optional):
                A function to apply to the predictions before they are passed to the
                pipeline. This is useful for preprocessing the predictions to match
                the expected input format of the pipeline. Defaults to a no-op function
                that returns the input unchanged.
            postprocessing_fn (optional):
                A function to apply to the metric scores after they are computed,
                taking the score to the postprocessed score along with its string
                representation. Defaults to x -> (100 * x, f"{x:.2%}").
        """
        super().__init__(
            name=name, pretty_name=pretty_name, postprocessing_fn=postprocessing_fn
        )
        self.pipeline_repo = pipeline_repo
        self.pipeline_file_name = pipeline_file_name
        self.pipeline_scoring_function = pipeline_scoring_function
        self.pipeline: "Pipeline | None" = None
        self.preprocessing_fn = preprocessing_fn

    def __call__(
        self,
        predictions: c.Sequence,
        references: c.Sequence,
        dataset: "Dataset",
        dataset_config: "DatasetConfig",
        benchmark_config: "BenchmarkConfig",
    ) -> float | None:
        """Calculate the metric score using the scikit-learn pipeline.

        Args:
            predictions:
                The model predictions.
            references:
                Not used, but required for consistency with the Metric interface.
            dataset:
                The dataset used for evaluation. This is only used in case any
                additional metadata is used to compute the metrics.
            dataset_config:
                The dataset configuration.
            benchmark_config:
                The benchmark configuration.

        Returns:
            The calculated metric score, or None if the score should be ignored.
        """
        if self.pipeline is None:
            self.pipeline = self._download_pipeline(
                cache_dir=benchmark_config.cache_dir
            )
        if self.preprocessing_fn is not None:
            predictions = self.preprocessing_fn(
                predictions=predictions, dataset=dataset
            )
        return self.pipeline_scoring_function(self.pipeline, predictions)

    def _download_pipeline(self, cache_dir: str) -> "Pipeline":
        """Download the scikit-learn pipeline from the given URL.

        Args:
            cache_dir:
                The directory to use for caching the downloaded pipeline.

        Returns:
            The downloaded scikit-learn pipeline.

        Raises:
            InvalidBenchmark:
                If the loading of the pipeline fails for any reason.
        """
        log(f"Loading pipeline from {self.pipeline_repo}...", level=logging.DEBUG)
        with no_terminal_output():
            folder_path = hf_hub.HfApi(
                token=unscramble("XbjeOLhwebEaSaDUMqqaPaPIhgOcyOfDpGnX_")
            ).snapshot_download(
                repo_id=self.pipeline_repo, repo_type="model", cache_dir=cache_dir
            )
        model_path = Path(folder_path, self.pipeline_file_name)
        try:
            with model_path.open(mode="rb") as f:
                pipeline = cloudpickle.load(f)
        except Exception as e:
            raise InvalidBenchmark(
                f"Failed to load pipeline from {self.pipeline_repo!r}: {e}"
            ) from e
        log(f"Successfully loaded pipeline: {pipeline}", level=logging.DEBUG)
        return pipeline


### European Values Metric ###


def european_values_preprocessing_fn(
    predictions: c.Sequence[int], dataset: "Dataset"
) -> c.Sequence[int]:
    """Preprocess the model predictions for the European Values metric.

    Args:
        predictions:
            The model predictions, a sequence of integers representing the predicted
            choices for each question.
        dataset:
            The dataset used for evaluation. This is only used in case any additional
            metadata is used to compute the metrics.

    Returns:
        The preprocessed model predictions, a sequence of integers representing the
        final predicted choices for each question after any necessary aggregation and
        mapping.

    Raises:
        AssertionError:
            If the number of predictions is not a multiple of 53, which is required
            for the European Values metric.
    """
    num_questions = 53
    num_phrasings_per_question = 5

    # Convert the predictions to integers
    integer_predictions: list[int] = []
    for prediction, idx_to_choice in zip(predictions, dataset["idx_to_choice"]):
        idx_to_choice = {
            int(idx): int(choice)
            for idx, choice in idx_to_choice.items()
            if choice is not None
        }
        if prediction not in idx_to_choice:
            raise InvalidBenchmark(
                f"The prediction {prediction} is not a valid index for the "
                f"question with choices {idx_to_choice}."
            )
        integer_prediction = idx_to_choice[prediction]
        integer_predictions.append(integer_prediction)

    assert len(predictions) % num_questions == 0, (
        f"The number of predictions ({len(predictions)}) is not a multiple of "
        f"{num_questions}, which is required for the European Values metric."
    )

    # When we are using the situational version of the dataset, there are 5 phrasings
    # for each question, so we need to aggregate the predictions by question, which we
    # do using majority voting.
    using_situational = len(predictions) == num_questions * num_phrasings_per_question
    if using_situational:
        # Reshape the predictions to a 2D array with `num_phrasings_per_question` rows
        # (one for each phrasing) and `num_questions` columns (one for each question).
        # The five phrasings for each question appear right after each other, e.g.,
        # (0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, ...)
        # Shape: (num_questions, num_phrasings_per_question)
        arr = np.array(
            [
                integer_predictions[i : i + num_phrasings_per_question]
                for i in range(0, len(predictions), num_phrasings_per_question)
            ]
        )

        # Double check that we reshaped the predictions correctly
        for idx, pred in enumerate(integer_predictions):
            assert arr[idx // 5, idx % 5] == pred, (
                f"Reshaped predictions do not match the original predictions at index "
                f"{idx}: {arr[idx // 5, idx % 5]} != {pred}."
            )

        # Use majority voting to get the final prediction for each question
        # Shape: (53,)
        arr = np.apply_along_axis(lambda x: np.bincount(x).argmax(), axis=1, arr=arr)  #  type: ignore[no-matching-overload]

        # Convert the array to a list
        integer_predictions = arr.tolist()

    # Some of the questions are categorical and we're only interested in whether the
    # model chooses a specific choice or not. This mapping takes the question index
    # to the choice value that we're interested in.
    question_choices = {
        0: 1,
        1: 5,
        3: 3,
        6: 1,
        15: 4,
        20: 2,
        47: 8,
        48: 7,
        49: 4,
        51: 4,
        52: 4,
    }

    # Map the predictions to the choices we're interested in
    integer_predictions = list(integer_predictions)
    for question_idx, choice in question_choices.items():
        integer_predictions[question_idx] = (
            1 if integer_predictions[question_idx] == choice else 0
        )

    return integer_predictions


def european_values_scoring_function(
    pipeline: "Pipeline", predictions: c.Sequence[int]
) -> float:
    """Scoring function for the European Values metric."""
    normalised_predictions = pipeline[0].transform([predictions])
    log_likelihoods = pipeline[1].transform(normalised_predictions)[0]
    score = sigmoid(pipeline[2].alpha_ * (log_likelihoods - pipeline[2].center_))
    return score.item()


european_values_metric = PipelineMetric(
    name="european_values",
    pretty_name="European Values",
    pipeline_repo="EuroEval/european-values-pipeline",
    pipeline_scoring_function=european_values_scoring_function,
    preprocessing_fn=european_values_preprocessing_fn,
)



"""Tests for the `data_loading` module."""

import os
from collections.abc import Generator
from functools import partial
from pathlib import Path

import pytest
from datasets import DatasetDict
from numpy.random import default_rng
from transformers.models.auto.tokenization_auto import AutoTokenizer

from euroeval.benchmark_modules.litellm import LiteLLMModel
from euroeval.constants import MAX_CONTEXT_LENGTH
from euroeval.data_loading import load_data, load_raw_data
from euroeval.data_models import BenchmarkConfig, DatasetConfig
from euroeval.dataset_configs import get_all_dataset_configs, get_dataset_config
from euroeval.enums import GenerativeType
from euroeval.generation_utils import apply_prompt, extract_few_shot_examples
from euroeval.tasks import RC


@pytest.fixture(scope="module")
def tokeniser_id() -> Generator[str, None, None]:
    """Fixture for the tokeniser ID."""
    yield "EuroEval/gemma-3-tokenizer"


class TestLoadData:
    """Tests for the `load_data` function."""

    @pytest.fixture(scope="class")
    def datasets(
        self, benchmark_config: BenchmarkConfig
    ) -> Generator[list[DatasetDict], None, None]:
        """A loaded dataset."""
        yield load_data(
            rng=default_rng(seed=4242),
            dataset_config=get_dataset_config(
                dataset_name="multi-wiki-qa-da",
                custom_datasets_file=Path("custom_datasets.py"),
            ),
            benchmark_config=benchmark_config,
        )

    def test_load_data_is_list_of_dataset_dicts(
        self, datasets: list[DatasetDict]
    ) -> None:
        """Test that the `load_data` function returns a list of `DatasetDict`."""
        assert isinstance(datasets, list)
        assert all(isinstance(d, DatasetDict) for d in datasets)

    def test_split_names_are_correct(self, datasets: list[DatasetDict]) -> None:
        """Test that the split names are correct."""
        assert all(set(d.keys()) == {"train", "val", "test"} for d in datasets)

    def test_number_of_iterations_is_correct(
        self, datasets: list[DatasetDict], benchmark_config: BenchmarkConfig
    ) -> None:
        """Test that the number of iterations is correct."""
        assert len(datasets) == benchmark_config.num_iterations

    def test_no_empty_examples(self, datasets: list[DatasetDict]) -> None:
        """Test that there are no empty examples in the datasets."""
        for dataset in datasets:
            for split in dataset.values():
                for feature in ["text", "tokens"]:
                    if feature in split.features:
                        assert all(len(x) > 0 for x in split[feature])


@pytest.mark.parametrize(
    argnames="dataset_config",
    argvalues=[
        dataset_config
        for dataset_config in get_all_dataset_configs(
            custom_datasets_file=Path("custom_datasets.py")
        ).values()
        if os.getenv("CHECK_DATASET") is not None
        and (
            dataset_config.name in os.environ["CHECK_DATASET"].split(",")
            or any(
                language.code in os.environ["CHECK_DATASET"].split(",")
                for language in dataset_config.languages
            )
            or "all" in os.environ["CHECK_DATASET"].split(",")
        )
    ],
    ids=lambda dc: dc.name,
)
class TestAllDatasets:
    """Tests that are run on all datasets."""

    def test_examples_in_official_datasets_are_not_too_long(
        self,
        dataset_config: DatasetConfig,
        benchmark_config: BenchmarkConfig,
        tokeniser_id: str,
    ) -> None:
        """Test that the examples are not too long in official datasets."""
        dummy_model_config = LiteLLMModel.get_model_config(
            model_id="model", benchmark_config=benchmark_config
        )
        tokeniser = AutoTokenizer.from_pretrained(tokeniser_id)
        dataset = load_raw_data(
            dataset_config=dataset_config, cache_dir=benchmark_config.cache_dir
        )

        for itr_idx in range(10):
            if "train" in dataset_config.splits:
                few_shot_examples = (
                    extract_few_shot_examples(
                        dataset=dataset,
                        dataset_config=dataset_config,
                        benchmark_config=benchmark_config,
                        itr_idx=itr_idx,
                    )
                    if not dataset_config.task.requires_zero_shot
                    else []
                )
            else:
                few_shot_examples = []
            for instruction_model in [True, False]:
                prepared_test = dataset["test"].map(
                    partial(
                        apply_prompt,
                        few_shot_examples=few_shot_examples,
                        model_config=dummy_model_config,
                        dataset_config=dataset_config,
                        generative_type=(
                            GenerativeType.INSTRUCTION_TUNED
                            if instruction_model
                            else GenerativeType.BASE
                        ),
                        always_populate_text_field=True,
                        tokeniser=tokeniser,
                    ),
                    batched=True,
                    load_from_cache_file=False,
                    keep_in_memory=True,
                )

                max_input_length = max(
                    len(tokeniser(prompt)["input_ids"])
                    for prompt in prepared_test["text"]
                )
                max_output_length = dataset_config.max_generated_tokens
                max_length = max_input_length + max_output_length

                assert max_length <= MAX_CONTEXT_LENGTH, (
                    f"Max length of {max_length:,} exceeds the maximum context length "
                    f"({MAX_CONTEXT_LENGTH:,}) for dataset {dataset_config.name} in "
                    f"iteration {itr_idx} and when instruction_model="
                    f"{instruction_model}."
                )

    def test_reading_comprehension_datasets_have_id_column(
        self, dataset_config: DatasetConfig, benchmark_config: BenchmarkConfig
    ) -> None:
        """Test that reading comprehension datasets have an ID column."""
        # Skip if the dataset is not a reading comprehension dataset
        if dataset_config.task != RC:
            pytest.skip(reason="Skipping test for non-reading comprehension dataset.")

        dataset = load_raw_data(
            dataset_config=dataset_config, cache_dir=benchmark_config.cache_dir
        )
        for split in dataset_config.splits:
            assert "id" in dataset[split].features, (
                f"Dataset {dataset_config.name} is a reading comprehension dataset but "
                f"the {split} split does not have an 'id' column."
            )



"""All the metrics used in EuroEval."""

from .bias import *  # noqa: F403
from .huggingface import *  # noqa: F403
from .llm_as_a_judge import *  # noqa: F403
from .pipeline import *  # noqa: F403
from .speed import *  # noqa: F403



"""All Swedish dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import SWEDISH
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, KNOW, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

SWEREC_CONFIG = DatasetConfig(
    name="swerec",
    pretty_name="SweReC",
    source="EuroEval/swerec-mini",
    task=SENT,
    languages=[SWEDISH],
)

SCALA_SV_CONFIG = DatasetConfig(
    name="scala-sv",
    pretty_name="ScaLA-sv",
    source="EuroEval/scala-sv",
    task=LA,
    languages=[SWEDISH],
)

SUC3_CONFIG = DatasetConfig(
    name="suc3",
    pretty_name="SUC3",
    source="EuroEval/suc3-mini",
    task=NER,
    languages=[SWEDISH],
)

MULTI_WIKI_QA_SV_CONFIG = DatasetConfig(
    name="multi-wiki-qa-sv",
    pretty_name="MultiWikiQA-sv",
    source="EuroEval/multi-wiki-qa-sv-mini",
    task=RC,
    languages=[SWEDISH],
)

SWEDN_CONFIG = DatasetConfig(
    name="swedn",
    pretty_name="SweDN",
    source="EuroEval/swedn-mini",
    task=SUMM,
    languages=[SWEDISH],
)

MMLU_SV_CONFIG = DatasetConfig(
    name="mmlu-sv",
    pretty_name="MMLU-sv",
    source="EuroEval/mmlu-sv-mini",
    task=KNOW,
    languages=[SWEDISH],
)

HELLASWAG_SV_CONFIG = DatasetConfig(
    name="hellaswag-sv",
    pretty_name="HellaSwag-sv",
    source="EuroEval/hellaswag-sv-mini",
    task=COMMON_SENSE,
    languages=[SWEDISH],
)

VALEU_SV_CONFIG = DatasetConfig(
    name="valeu-sv",
    pretty_name="VaLEU-sv",
    source="EuroEval/european-values-sv",
    task=EUROPEAN_VALUES,
    languages=[SWEDISH],
    splits=["test"],
    bootstrap_samples=False,
    _instruction_prompt="{text}",
)


### Unofficial datasets ###

SCHIBSTED_SV_CONFIG = DatasetConfig(
    name="schibsted-sv",
    pretty_name="Schibsted-sv",
    source="EuroEval/schibsted-article-summaries-sv",
    task=SUMM,
    languages=[SWEDISH],
    unofficial=True,
)

ARC_SV_CONFIG = DatasetConfig(
    name="arc-sv",
    pretty_name="ARC-sv",
    source="EuroEval/arc-sv-mini",
    task=KNOW,
    languages=[SWEDISH],
    unofficial=True,
)

BELEBELE_SV_CONFIG = DatasetConfig(
    name="belebele-sv",
    pretty_name="Belebele-sv",
    source="EuroEval/belebele-sv-mini",
    task=MCRC,
    languages=[SWEDISH],
    unofficial=True,
)

SCANDIQA_SV_CONFIG = DatasetConfig(
    name="scandiqa-sv",
    pretty_name="ScandiQA-sv",
    source="EuroEval/scandiqa-sv-mini",
    task=RC,
    languages=[SWEDISH],
    unofficial=True,
)

GOLDENSWAG_SV_CONFIG = DatasetConfig(
    name="goldenswag-sv",
    pretty_name="GoldenSwag-sv",
    source="EuroEval/goldenswag-sv-mini",
    task=COMMON_SENSE,
    languages=[SWEDISH],
    unofficial=True,
)

WINOGRANDE_SV_CONFIG = DatasetConfig(
    name="winogrande-sv",
    pretty_name="Winogrande-sv",
    source="EuroEval/winogrande-sv",
    task=COMMON_SENSE,
    languages=[SWEDISH],
    _labels=["a", "b"],
    unofficial=True,
)

SKOLPROV_CONFIG = DatasetConfig(
    name="skolprov",
    pretty_name="Skolprov",
    source="EuroEval/skolprov",
    task=KNOW,
    languages=[SWEDISH],
    unofficial=True,
)

SWEDISH_FACTS_CONFIG = DatasetConfig(
    name="swedish-facts",
    pretty_name="Swedish Facts",
    source="EuroEval/swedish-facts",
    task=KNOW,
    languages=[SWEDISH],
    unofficial=True,
)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the Jentoft linguistic acceptability dataset."""

import logging

import pandas as pd
from datasets import Dataset, DatasetDict, Split
from huggingface_hub import HfApi

from .constants import MAX_NUM_CHARS_IN_DOCUMENT, MIN_NUM_CHARS_IN_DOCUMENT  # noqa

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("create_jentoft")


def main() -> None:
    """Create the Jentoft linguistic acceptability dataset and upload to HF Hub."""
    dataset_urls = {
        "val": (
            "https://raw.githubusercontent.com/matias-jjj/matias_master/refs/heads/"
            "main/norgec_dataset/ask_exp_dev.tsv"
        ),
        "test": (
            "https://raw.githubusercontent.com/matias-jjj/matias_master/refs/heads/"
            "main/norgec_dataset/ask_exp_test.tsv"
        ),
        "train": (
            "https://raw.githubusercontent.com/matias-jjj/matias_master/refs/heads/"
            "main/norgec_dataset/ask_exp_train.tsv"
        ),
    }

    full_val_df = prepare_dataset(dataset_url=dataset_urls["val"])
    full_test_df = prepare_dataset(dataset_url=dataset_urls["test"])
    full_train_df = prepare_dataset(dataset_url=dataset_urls["train"])

    # Remove overlapping texts between splits
    full_train_df, full_val_df, full_test_df = remove_overlapping_samples(
        full_train_df=full_train_df, full_val_df=full_val_df, full_test_df=full_test_df
    )

    assert isinstance(full_train_df, pd.DataFrame)
    assert isinstance(full_val_df, pd.DataFrame)
    assert isinstance(full_test_df, pd.DataFrame)

    val_size = 256
    test_size = 2048
    train_size = 1024

    val_df = sample_dataset(full_df=full_val_df, size=val_size)
    test_df = sample_dataset(full_df=full_test_df, size=test_size)
    train_df = sample_dataset(full_df=full_train_df, size=train_size)

    # Only work with samples where the document is not very large or small
    # We do it after we have made the splits to ensure that the dataset is minimally
    # affected.
    new_train_df = train_df.copy()
    new_train_df["text_len"] = new_train_df.text.str.len()
    new_train_df = new_train_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_val_df = val_df.copy()
    new_val_df["text_len"] = new_val_df.text.str.len()
    new_val_df = new_val_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_test_df = test_df.copy()
    new_test_df["text_len"] = new_test_df.text.str.len()
    new_test_df = new_test_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(
                new_train_df, split=Split.TRAIN, preserve_index=False
            ),
            "val": Dataset.from_pandas(
                new_val_df, split=Split.VALIDATION, preserve_index=False
            ),
            "test": Dataset.from_pandas(
                new_test_df, split=Split.TEST, preserve_index=False
            ),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/jentoft-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


def prepare_dataset(dataset_url: str) -> pd.DataFrame:
    """Prepare the dataset from a URL.

    Args:
        dataset_url:
            The URL to the dataset.

    Returns:
        The prepared dataset.
    """
    df = pd.read_csv(dataset_url, sep="\t", on_bad_lines="skip")

    # Make label column
    df["label"] = df["ERROR"].apply(
        lambda x: "correct" if x == "no_error" else "incorrect"
    )

    # Rename the columns
    df = df.rename(columns={"SOURCE": "text"})

    # Only keep relevant columns
    df = df.loc[["text", "label"]]

    # Remove text duplicates
    df = df.drop_duplicates(subset=["text"])

    # Remove samples with with 5 or fewer words
    df = df[df["text"].str.split().apply(len) > 5]
    return df


def sample_dataset(full_df: pd.DataFrame, size: int) -> pd.DataFrame:
    """Sample a dataset with an equal number of 'correct' and 'incorrect' samples.

    Args:
        full_df:
            The full dataset.
        size:
            The size of the sampled dataset.

    Returns:
        The sampled dataset.
    """
    half_size = size // 2
    assert (full_df["label"] == "correct").sum() >= half_size, (
        "Not enough 'correct' samples"
    )

    correct_df = full_df[full_df["label"] == "correct"].sample(
        n=half_size, random_state=4242
    )
    incorrect_df = full_df[full_df["label"] == "incorrect"].sample(
        n=half_size, random_state=4242
    )
    concatenated_df = pd.concat(objs=[correct_df, incorrect_df])
    assert isinstance(concatenated_df, pd.DataFrame)
    return concatenated_df


def remove_overlapping_samples(
    full_train_df: pd.DataFrame, full_val_df: pd.DataFrame, full_test_df: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Ensure that a sample is only present in one split.

    Args:
        full_train_df:
            The full training dataset.
        full_val_df:
            The full validation dataset.
        full_test_df:
            The full test dataset.

    Returns:
        The cleaned training, validation, and test datasets.
    """
    train_texts = set(full_train_df["text"])
    val_texts = set(full_val_df["text"])
    test_texts = set(full_test_df["text"])

    # Find overlaps
    train_val_overlap = train_texts.intersection(val_texts)
    train_test_overlap = train_texts.intersection(test_texts)
    val_test_overlap = val_texts.intersection(test_texts)

    # Print overlap statistics
    logger.info(
        f"Overlapping texts between train and validation: {len(train_val_overlap)}"
    )
    logger.info(f"Overlapping texts between train and test: {len(train_test_overlap)}")
    logger.info(
        f"Overlapping texts between validation and test: {len(val_test_overlap)}"
    )

    # Remove overlapping texts from validation and test sets, prioritize train >
    # validation > test
    clean_val_df = full_val_df.loc[~full_val_df["text"].isin(list(train_texts))]
    clean_test_df = full_test_df.loc[
        ~full_test_df["text"].isin(list(train_texts))
        & ~full_test_df["text"].isin(list(val_texts))
    ]

    # The training set remains unchanged
    clean_train_df = full_train_df

    # Verify no more overlaps
    assert len(set(clean_train_df["text"]).intersection(set(clean_val_df["text"]))) == 0
    assert (
        len(set(clean_train_df["text"]).intersection(set(clean_test_df["text"]))) == 0
    )
    assert len(set(clean_val_df["text"]).intersection(set(clean_test_df["text"]))) == 0

    return clean_train_df, clean_val_df, clean_test_df


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn<1.6.0",
# ]
# ///

"""Create the Exam-et dataset and upload it to the HF Hub."""

from collections import Counter

import pandas as pd
from datasets import Dataset, DatasetDict, load_dataset
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split

from .constants import (
    CHOICES_MAPPING,
    MAX_NUM_CHARS_IN_INSTRUCTION,
    MAX_NUM_CHARS_IN_OPTION,
    MAX_REPETITIONS,
    MIN_NUM_CHARS_IN_INSTRUCTION,
    MIN_NUM_CHARS_IN_OPTION,
)


def main() -> None:
    """Create the Exam-et dataset and upload it to the HF Hub."""
    repo_id = "TalTechNLP/exam_et"
    target_repo_id = "EuroEval/exam-et"

    # Get all the subsets of the dataset
    api = HfApi(token=True)
    repo_info = api.repo_info(repo_id=repo_id, repo_type="dataset")
    subsets = [config["config_name"] for config in repo_info.card_data.configs]

    # Download all subsets and merge them
    rename_mapping = {
        "küsimus": "instruction",
        "vastusevariandid": "choices",
        "vastus": "label",
    }
    dfs: list[pd.DataFrame] = []
    for subset in subsets:
        ds = load_dataset(path=repo_id, name=subset, split="train", token=True)
        assert isinstance(ds, Dataset), f"Expected Dataset, got {type(ds)}"
        df = ds.to_pandas()
        assert isinstance(df, pd.DataFrame), f"Expected DataFrame, got {type(df)}"
        df.rename(columns=rename_mapping, inplace=True)
        df["category"] = subset
        df = df.loc[:, ["instruction", "choices", "label", "category"]]
        dfs.append(df)
    df = pd.concat(dfs, ignore_index=True)

    # Remove the samples with overly short or long texts
    df = df.loc[
        (df.instruction.str.len() >= MIN_NUM_CHARS_IN_INSTRUCTION)
        & (df.instruction.str.len() <= MAX_NUM_CHARS_IN_INSTRUCTION)
        & df.choices.map(
            lambda endings: min(len(ending) for ending in endings)
            >= MIN_NUM_CHARS_IN_OPTION
            and max(len(ending) for ending in endings) <= MAX_NUM_CHARS_IN_OPTION
        )
    ]

    def is_repetitive(text: str) -> bool:
        """Return True if the text is repetitive."""
        max_repetitions = max(Counter(text.split()).values())
        return max_repetitions > MAX_REPETITIONS

    # Remove overly repetitive samples
    df = df.loc[
        ~df.instruction.apply(is_repetitive)
        & ~df.choices.map(
            lambda endings: any(is_repetitive(ending) for ending in endings)
        )
    ]

    # Make a `text` column with all the options in it
    label_str = "abcdefghijklmnopqrstuvwxyz"
    df["text"] = [
        row.instruction.replace("\n", " ").strip()
        + "\n"
        + f"{CHOICES_MAPPING['et']}:\n"
        + "\n".join(
            f"{char}. {option.strip()}" for char, option in zip(label_str, row.choices)
        )
        for _, row in df.iterrows()
    ]
    df["label"] = df.label.map(lambda x: label_str[x])

    # Keep only the relevant columns
    df = df[["text", "label", "category"]]

    train_size = 512
    val_size = 64
    test_size = 896
    assert len(df) == train_size + val_size + test_size, (
        f"Expected {train_size + val_size + test_size:,} samples, got {len(df):,}"
    )

    # Create splits, stratifiying by category
    train_df, valtest_df = train_test_split(
        df, train_size=train_size, random_state=4242, stratify=df.category
    )
    val_df, test_df = train_test_split(
        valtest_df, test_size=test_size, random_state=4242, stratify=valtest_df.category
    )
    assert isinstance(train_df, pd.DataFrame), (
        f"Expected DataFrame, got {type(train_df)}"
    )
    assert isinstance(val_df, pd.DataFrame), f"Expected DataFrame, got {type(val_df)}"
    assert isinstance(test_df, pd.DataFrame), f"Expected DataFrame, got {type(test_df)}"

    # Shuffle the splits
    train_df = train_df.sample(frac=1, random_state=4242).reset_index(drop=True)
    val_df = val_df.sample(frac=1, random_state=4242).reset_index(drop=True)
    test_df = test_df.sample(frac=1, random_state=4242).reset_index(drop=True)

    # Convert to DatasetDict
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df),
            "val": Dataset.from_pandas(val_df),
            "test": Dataset.from_pandas(test_df),
        }
    )

    api.delete_repo(repo_id=target_repo_id, repo_type="dataset", missing_ok=True)
    dataset.push_to_hub(target_repo_id, private=True)


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the FoNE-mini NER dataset and upload it to the HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi


def main() -> None:
    """Create the FoNE-mini NER dataset and uploads it to the HF Hub."""
    # Define dataset ID
    repo_id = "vesteinn/sosialurin-faroese-ner"

    # Download the dataset
    dataset = load_dataset(path=repo_id, token=True, split="train")
    assert isinstance(dataset, Dataset)

    # Convert the dataset to a dataframe
    df = dataset.to_pandas()
    assert isinstance(df, pd.DataFrame)

    # Drop all columns except for `tokens` and `ner_tags`
    columns_to_drop = [col for col in df.columns if col not in ["tokens", "ner_tags"]]
    df.drop(columns=columns_to_drop, inplace=True)

    # Add a `text` column
    df["text"] = df["tokens"].map(lambda tokens: " ".join(tokens))

    # Rename `ner_tags` to `labels`
    df.rename(columns={"ner_tags": "labels"}, inplace=True)

    # Convert the NER tags from IDs to strings
    ner_conversion_dict = {
        0: "O",  # Original: B-Date
        1: "B-LOC",  # Original: B-Location
        2: "B-MISC",  # Original: B-Miscellaneous
        3: "O",  # Original: B-Money
        4: "B-ORG",  # Original: B-Organization
        5: "O",  # Original: B-Percent
        6: "B-PER",  # Original: B-Person
        7: "O",  # Original: B-Time
        8: "O",  # Original: I-Date
        9: "I-LOC",  # Original: I-Location
        10: "I-MISC",  # Original: I-Miscellaneous
        11: "O",  # Original: I-Money
        12: "I-ORG",  # Original: I-Organization
        13: "O",  # Original: I-Percent
        14: "I-PER",  # Original: I-Person
        15: "O",  # Original: I-Time
        16: "O",  # Original: O
    }
    df["labels"] = df["labels"].map(
        lambda ner_tags: [ner_conversion_dict[ner_tag] for ner_tag in ner_tags]
    )

    for token_list, ner_tag_list in zip(df["tokens"], df["labels"]):
        # Sanity check that the number of tokens and named entity tags are equal
        assert len(token_list) == len(ner_tag_list), (
            "The number of tokens and named entity tags are not equal."
        )

        # Fix invalid I-tags
        invalid_i_ner_tags = [
            ner_tag
            for token_idx, ner_tag in enumerate(ner_tag_list)
            if ner_tag.startswith("I-")
            and ner_tag_list[token_idx - 1] not in {f"B-{ner_tag[2:]}", ner_tag}
        ]
        while invalid_i_ner_tags:
            for invalid_i_ner_tag in invalid_i_ner_tags:
                ner_tag_list[ner_tag_list.index(invalid_i_ner_tag)] = (
                    f"B-{invalid_i_ner_tag[2:]}"
                )
            invalid_i_ner_tags = [
                ner_tag
                for token_idx, ner_tag in enumerate(ner_tag_list)
                if ner_tag.startswith("I-")
                and ner_tag_list[token_idx - 1] not in {f"B-{ner_tag[2:]}", ner_tag}
            ]

        # Sanity check that all I-tags are valid
        assert not invalid_i_ner_tags, (
            f"The following I- tags are invalid: {invalid_i_ner_tags}."
        )

    # Create validation split
    val_size = 256
    val_df = df.sample(
        n=val_size,
        random_state=4242,
        weights=[5.0 if len(set(labels)) > 1 else 1.0 for labels in df["labels"]],
    )

    # Create test split
    test_size = 2048
    filtered_df = df.loc[~df.index.isin(val_df.index)]
    test_df = filtered_df.sample(
        n=test_size,
        random_state=4242,
        weights=[
            5.0 if len(set(labels)) > 1 else 1.0 for labels in filtered_df["labels"]
        ],
    )

    # Create train split
    train_size = 1024
    filtered_df = filtered_df.loc[~filtered_df.index.isin(test_df.index)]
    train_df = filtered_df.sample(
        n=train_size,
        random_state=4242,
        weights=[
            5.0 if len(set(labels)) > 1 else 1.0 for labels in filtered_df["labels"]
        ],
    )

    # Reset the index
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/fone-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



"""Callbacks for the Hugging Face Trainer."""

import sys
import typing as t
from collections.abc import Sized

from tqdm.auto import tqdm
from transformers.trainer_callback import ProgressCallback

from .logging_utils import get_pbar

if t.TYPE_CHECKING:
    from torch.utils.data import DataLoader
    from transformers.trainer_callback import TrainerControl, TrainerState
    from transformers.training_args import TrainingArguments


class NeverLeaveProgressCallback(ProgressCallback):
    """Progress callback which never leaves the progress bar."""

    def __init__(self, max_str_len: int = 100) -> None:
        """Initialise the callback."""
        super().__init__(max_str_len=max_str_len)
        self.training_bar: tqdm | None = None
        self.prediction_bar: tqdm | None = None

    def on_train_begin(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        **kwargs: str,
    ) -> None:
        """Callback actions when training begins."""
        if state.is_local_process_zero:
            desc = "Finetuning model"
            self.training_bar = get_pbar(
                total=None, desc=desc, disable=hasattr(sys, "_called_from_test")
            )
        self.current_step = 0

    def on_step_end(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        **kwargs: str,
    ) -> None:
        """Callback actions when a training step ends."""
        if state.is_local_process_zero and self.training_bar is not None:
            self.training_bar.update(state.global_step - self.current_step)
            self.current_step = state.global_step

    def on_prediction_step(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        eval_dataloader: "DataLoader | None" = None,
        **kwargs: str,
    ) -> None:
        """Callback actions when a prediction step ends."""
        if eval_dataloader is None:
            return
        correct_dtype = isinstance(eval_dataloader.dataset, Sized)
        if state.is_local_process_zero and correct_dtype:
            if self.prediction_bar is None:
                desc = "Evaluating model"
                self.prediction_bar = get_pbar(
                    total=len(eval_dataloader),
                    desc=desc,
                    disable=hasattr(sys, "_called_from_test"),
                )
            self.prediction_bar.update(1)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the MMS sentiment datasets and upload to HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi
from sklearn.utils import resample

LANGUAGES = ["sr", "hr", "bs", "sq"]


def main() -> None:
    """Create the MMS sentiment datasets and upload to HF Hub."""
    # Define the dataset repository
    repo_id = "Brand24/mms"

    # Download the dataset
    dataset = load_dataset(path=repo_id)
    assert isinstance(dataset, DatasetDict)

    # Get the train split (MMS only has train split)
    df_all_languages = dataset["train"].to_pandas()
    assert isinstance(df_all_languages, pd.DataFrame)

    for language in LANGUAGES:
        # Filter based on language
        df = df_all_languages[df_all_languages["language"] == language].reset_index(
            drop=True
        )
        assert isinstance(df, pd.DataFrame)

        # Map numeric labels to string labels
        # 0: negative, 1: neutral, 2: positive
        df["label"] = df["label"].map(
            lambda x: {0: "negative", 1: "neutral", 2: "positive"}[x]
        )

        # Remove duplicates
        df = df.drop_duplicates().reset_index(drop=True)

        # Create a uniform distribution based on the original dataset and label columns
        df_uniform_original_dataset = create_uniform_distribution(
            df=df, column="original_dataset"
        )
        df_uniform_label = create_uniform_distribution(
            df=df_uniform_original_dataset, column="label"
        )
        df_remaining = df[~df.index.isin(df_uniform_label.index)]
        assert isinstance(df_remaining, pd.DataFrame)
        df_remaining_uniform_label = create_uniform_distribution(
            df=df_remaining, column="label"
        )
        df = df_uniform_label

        # Define split sizes
        test_size = 2048
        val_size = 256
        train_size = 1024

        n_missing_samples = test_size + val_size + train_size - len(df)

        if n_missing_samples > 0:
            df_additional = df_remaining_uniform_label.sample(
                n=n_missing_samples, random_state=4242
            )
            df = pd.concat([df, df_additional])

        # Shuffle the dataframe
        df = df.sample(frac=1, random_state=4242).reset_index(drop=True)

        # Keep only text and label columns
        df = df[["text", "label"]]

        # Create splits
        test_df = df.iloc[:test_size].reset_index(drop=True)
        val_df = df.iloc[test_size : test_size + val_size].reset_index(drop=True)
        train_df = df.iloc[
            test_size + val_size : test_size + val_size + train_size
        ].reset_index(drop=True)

        # Create dataset dictionary with custom splits
        dataset_dict = DatasetDict(
            {
                "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
                "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
                "test": Dataset.from_pandas(test_df, split=Split.TEST),
            }
        )

        # Push the dataset to the Hugging Face Hub
        dataset_id = f"EuroEval/mms-{language}-mini"
        HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
        dataset_dict.push_to_hub(dataset_id, private=True)


def create_uniform_distribution(
    df: pd.DataFrame, column: str = "label", random_state: int = 4242
) -> pd.DataFrame:
    """Create a sampled dataset with a uniform distribution for the given column.

    Args:
        df: The input dataframe.
        column: The name of the column to create a uniform label distribution for
        random_state: The random state for reproducibility.

    Returns:
        A dataframe with a uniform label distribution.
    """
    # Separate each class
    classes = df[column].unique()
    class_dfs = [df[df[column] == label] for label in classes]

    # Find the size of the smallest class
    min_size = min(len(class_df) for class_df in class_dfs)

    # Resample each class to the size of the smallest class
    resampled_dfs = [
        resample(class_df, replace=False, n_samples=min_size, random_state=random_state)
        for class_df in class_dfs
    ]

    # Combine the resampled dataframes (keep original indices!)
    balanced_df = pd.concat(resampled_dfs, ignore_index=False)

    return balanced_df


if __name__ == "__main__":
    main()



"""Templates for the Sentiment Analysis task."""

import typing as t

from ..data_models import PromptConfig
from ..languages import (
    ALBANIAN,
    BOSNIAN,
    BULGARIAN,
    CATALAN,
    CROATIAN,
    CZECH,
    DANISH,
    DUTCH,
    ENGLISH,
    ESTONIAN,
    FAROESE,
    FINNISH,
    FRENCH,
    GERMAN,
    GREEK,
    HUNGARIAN,
    ICELANDIC,
    ITALIAN,
    LATVIAN,
    LITHUANIAN,
    NORWEGIAN,
    NORWEGIAN_BOKMÅL,
    NORWEGIAN_NYNORSK,
    POLISH,
    PORTUGUESE,
    ROMANIAN,
    SERBIAN,
    SLOVAK,
    SLOVENE,
    SPANISH,
    SWEDISH,
    UKRAINIAN,
)

if t.TYPE_CHECKING:
    from ..languages import Language

SENT_TEMPLATES: dict["Language", PromptConfig] = {
    ALBANIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitive", neutral="neutrale", negative="negative"
        ),
        default_prompt_prefix="Më poshtë janë dokumentet dhe ndjenjat e tyre, të cilat "
        "mund të jenë {labels_str}.",
        default_prompt_template="Dokument: {text}\nNdjenja: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasifikoni ndjenjën në "
        "dokument. Përgjigjuni vetëm me {labels_str}, dhe asgjë tjetër.",
    ),
    BOSNIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitivno", neutral="neutralno", negative="negativno"
        ),
        default_prompt_prefix="Slijede dokumenti i njihova osjetila, koja mogu biti "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nOsjetilo: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasificirajte osjećaj u "
        "dokumentu. Odgovorite samo s {labels_str}, i ništa drugo.",
    ),
    BULGARIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="позитивен", neutral="неутрален", negative="негативен"
        ),
        default_prompt_prefix="Следват документи и техният сентимент, който може да "
        "бъде{labels_str}.",
        default_prompt_template="Документ: {text}\nСентимент: {label}",
        default_instruction_prompt="Документ: {text}\n\nКласифицирайте сентимента в "
        "документа. Отговорете с {labels_str}, и нищо друго.",
    ),
    CATALAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiu", neutral="neutral", negative="negatiu"
        ),
        default_prompt_prefix="Els documents següents i el seu sentiment, que pot "
        "ser {labels_str}.",
        default_prompt_template="Document: {text}\nSentiment: {label}",
        default_instruction_prompt="Document: {text}\n\nClassifiqueu el sentiment en "
        "el document. Contesteu només amb {labels_str}, i res més.",
    ),
    DANISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="neutral", negative="negativ"
        ),
        default_prompt_prefix="Følgende er dokumenter og deres sentiment, som kan være "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassificer sentimentet i "
        "dokumentet. Svar kun med {labels_str}, og intet andet.",
    ),
    CROATIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitivno", neutral="neutralno", negative="negativno"
        ),
        default_prompt_prefix="Slijede dokumenti i njihova osjetila, koja mogu biti "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nOsjetilo: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasificirajte osjećaj u "
        "dokumentu. Odgovorite samo s {labels_str}, i ništa drugo.",
    ),
    CZECH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitivní", neutral="neutrální", negative="negativní"
        ),
        default_prompt_prefix="Následují dokumenty a jejich sentiment, který může být "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasifikujte sentiment v "
        "dokumentu. Odpovězte pouze s {labels_str}, a nic jiného.",
    ),
    GERMAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="neutral", negative="negativ"
        ),
        default_prompt_prefix="Nachfolgend finden Sie Dokumente und ihre Bewertung, "
        "die {labels_str} sein kann.",
        default_prompt_template="Dokument: {text}\nStimmung: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassifizieren Sie die "
        "Stimmung im Dokument. Antworten Sie mit {labels_str}, und nichts anderes.",
    ),
    GREEK: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="θετικό", neutral="ουδέτερο", negative="αρνητικό"
        ),
        default_prompt_prefix="Τα ακόλουθα είναι έγγραφα και το συναίσθημά τους, "
        "το οποίο μπορεί να είναι {labels_str}.",
        default_prompt_template="Έγγραφο: {text}\nΣυναίσθημα: {label}",
        default_instruction_prompt="Έγγραφο: {text}\n\nΤαξινομήστε το συναίσθημα "
        "στο έγγραφο. Απαντήστε με {labels_str}, και τίποτα άλλο.",
    ),
    HUNGARIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitív", neutral="semleges", negative="negatív"
        ),
        default_prompt_prefix="Az alábbiak dokumentumok és érzelmük, ami lehet "
        "{labels_str}.",
        default_prompt_template="Dokumentum: {text}\nÉrzelem: {label}",
        default_instruction_prompt="Dokumentum: {text}\n\nOsztályozza az érzelmet a "
        "dokumentumban. Válaszoljon {labels_str}, és semmi mással.",
    ),
    ENGLISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positive", neutral="neutral", negative="negative"
        ),
        default_prompt_prefix="The following are documents and their sentiment, which "
        "can be {labels_str}.",
        default_prompt_template="Document: {text}\nSentiment: {label}",
        default_instruction_prompt="Document: {text}\n\nClassify the sentiment in the "
        "document. Answer with {labels_str}, and nothing else.",
    ),
    SPANISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positivo", neutral="neutral", negative="negativo"
        ),
        default_prompt_prefix="A continuación se muestran los documentos y su "
        "sentimiento, que puede ser {labels_str}.",
        default_prompt_template="Documento: {text}\nSentimiento: {label}",
        default_instruction_prompt="Documento: {text}\n\nClasifica el sentimiento del "
        "documento. Responde con {labels_str}, y nada más.",
    ),
    ESTONIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiivne", neutral="neutraalne", negative="negatiivne"
        ),
        default_prompt_prefix="Järgmised on dokumendid ja nende meelestatus, "
        "mis võib olla {labels_str}.",
        default_prompt_template="Dokument: {text}\nMeelestatus: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassifitseeri dokument "
        "meelestatuse järgi. Võimalikud vastused: {labels_str}. Muud vastused "
        "ei ole lubatud.",
    ),
    POLISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozytywny", neutral="neutralny", negative="negatywny"
        ),
        default_prompt_prefix="Poniżej znajdują się dokumenty i ich sentyment, który "
        "może być {labels_str}.",
        default_prompt_template="Dokument: {text}\nSentyment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasyfikuj sentyment w "
        "dokumencie. Odpowiedz jednym słowem: {labels_str}.",
    ),
    PORTUGUESE: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positivo", neutral="neutro", negative="negativo"
        ),
        default_prompt_prefix="Abaixo encontras documentos e os seus "
        "sentimentos correspondentes, que podem ser {labels_str}.",
        default_prompt_template="Documento: {text}\nSentimento: {label}",
        default_instruction_prompt="Documento: {text}\n\nClassifica o "
        "sentimento do documento. Responde apenas com {labels_str}.",
    ),
    FINNISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiivinen", neutral="neutrali", negative="negatiivinen"
        ),
        default_prompt_prefix="Seuraavassa on arvosteluja ja niiden tunnesävy, joka "
        "voi olla {labels_str}.",
        default_prompt_template="Teksti: {text}\nTunnesävy: {label}",
        default_instruction_prompt="Teksti: {text}\n\nLuokittele arvostelun tunnesävy. "
        "Vastaa vain {labels_str}, ei muuta.",
    ),
    FAROESE: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positivt", neutral="neutralt", negative="negativt"
        ),
        default_prompt_prefix="Niðanfyri eru skjøl og teirra kenslur, sum kunnu vera "
        "{labels_str}.",
        default_prompt_template="Skjal: {text}\nKensla: {label}",
        default_instruction_prompt="Skjal: {text}\n\nFlokka kensluna í skjalinum. "
        "Svara við {labels_str}, og einki annað.",
    ),
    FRENCH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positif", neutral="neutre", negative="négatif"
        ),
        default_prompt_prefix="Les documents suivants sont accompagnés de leur "
        "sentiment, qui peut être {labels_str}.",
        default_prompt_template="Document: {text}\nSentiment: {label}",
        default_instruction_prompt="Document: {text}\n\nClassez le sentiment dans le "
        "document. Répondez par {labels_str}, et rien d'autre.",
    ),
    ICELANDIC: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="jákvætt", neutral="hlutlaust", negative="neikvætt"
        ),
        default_prompt_prefix="Hér fyrir neðan eru textabrot ásamt lyndisgildi þeirra "
        "sem getur verið 'jákvætt', 'hlutlaust' eða 'neikvætt'.",
        default_prompt_template="Textabrot: {text}\nViðhorf: {label}",
        default_instruction_prompt="Textabrot: {text}\n\nGreindu lyndið í "
        "textabrotinu. Svaraðu með {labels_str}, og ekkert annað.",
    ),
    ITALIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positivo", neutral="neutro", negative="negativo"
        ),
        default_prompt_prefix="Di seguito sono riportati i documenti e il loro "
        "sentiment, che può essere {labels_str}.",
        default_prompt_template="Documento: {text}\nSentimento: {label}",
        default_instruction_prompt="Documento: {text}\n\nClassificare il sentiment del "
        "documento. Rispondere con {labels_str}, e nient'altro.",
    ),
    LITHUANIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="teigiamas", neutral="neutralus", negative="neigiamas"
        ),
        default_prompt_prefix="Toliau pateikti dokumentai ir jų nuotaika, kuri "
        "gali būti {labels_str}.",
        default_prompt_template="Dokumentas: {text}\nNuotaika: {label}",
        default_instruction_prompt="Dokumentas: {text}\n\nKlasifikuokite nuotaiką "
        "dokumente. Atsakykite su {labels_str}, ir nieko kito.",
    ),
    LATVIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitīvs", neutral="neitrāls", negative="negatīvs"
        ),
        default_prompt_prefix="Tālāk ir dokumenti un to noskaņojums, kas var būt "
        "{labels_str}.",
        default_prompt_template="Dokuments: {text}\nNoskaņojums: {label}",
        default_instruction_prompt="Dokuments: {text}\n\nKlasificējiet noskaņojumu "
        "dokumentā. Atbildiet ar {labels_str}, un neko citu.",
    ),
    NORWEGIAN_BOKMÅL: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="nøytral", negative="negativ"
        ),
        default_prompt_prefix="Her følger dokumenter og deres sentiment, som kan være "
        "{labels_str}",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassifiser følelsen i "
        "teksten. Svar med {labels_str}, og ikke noe annet.",
    ),
    DUTCH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positief", neutral="neutraal", negative="negatief"
        ),
        default_prompt_prefix="Hieronder volgen documenten en hun sentiment, dat "
        "{labels_str} kan zijn.",
        default_prompt_template="Document: {text}\nSentiment: {label}",
        default_instruction_prompt="Document: {text}\n\nClassificeer het sentiment in "
        "het document. Antwoord met {labels_str}, en verder niets.",
    ),
    NORWEGIAN_NYNORSK: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="nøytral", negative="negativ"
        ),
        default_prompt_prefix="Her følger dokumenter og deres sentiment, som kan være "
        "{labels_str}",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassifiser følelsen i "
        "teksten. Svar med {labels_str}, og ikke noe annet.",
    ),
    NORWEGIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="nøytral", negative="negativ"
        ),
        default_prompt_prefix="Her følger dokumenter og deres sentiment, som kan være "
        "{labels_str}",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassifiser følelsen i "
        "teksten. Svar med {labels_str}, og ikke noe annet.",
    ),
    ROMANIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitiv", neutral="neutru", negative="negativ"
        ),
        default_prompt_prefix="Urmează documentele și sentimentul acestora, care "
        "poate fi {labels_str}.",
        default_prompt_template="Document: {text}\nSentiment: {label}",
        default_instruction_prompt="Document: {text}\n\nClasificați sentimentul "
        "documentului. Răspundeți cu {labels_str}, și nimic altceva.",
    ),
    SLOVAK: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitívne", neutral="neutrálne", negative="negatívne"
        ),
        default_prompt_prefix="Nižšie sú dokumenty a ich sentiment, ktorý môže byť "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasifikujte pocit v "
        "dokumente. Odpovedzte so {labels_str}, a nič iné.",
    ),
    SLOVENE: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitivno", neutral="nevtralno", negative="negativno"
        ),
        default_prompt_prefix="Spodaj so dokumenti in njihov sentiment, ki je lahko "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasificirajte sentiment v "
        "dokumentu. Odgovorite z {labels_str}, in nič drugega.",
    ),
    SERBIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="pozitivan", neutral="neutralan", negative="negativan"
        ),
        default_prompt_prefix="U nastavku su dokumenti i njihov sentiment, koji može "
        "biti {labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlasifikujte sentiment u "
        "dokumentu. Odgovorite sa {labels_str}, i ništa drugo.",
    ),
    SWEDISH: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="positiv", neutral="neutral", negative="negativ"
        ),
        default_prompt_prefix="Nedan följer dokument och deras sentiment, som kan vara "
        "{labels_str}.",
        default_prompt_template="Dokument: {text}\nSentiment: {label}",
        default_instruction_prompt="Dokument: {text}\n\nKlassificera känslan i "
        "dokumentet. Svara med {labels_str}, och inget annat.",
    ),
    UKRAINIAN: PromptConfig(
        default_prompt_label_mapping=dict(
            positive="позитивний", neutral="нейтральний", negative="негативний"
        ),
        default_prompt_prefix="Нижче наведені документи і їх настрій, який може "
        "бути {labels_str}.",
        default_prompt_template="Документ: {text}\nНастрій: {label}",
        default_instruction_prompt="Документ: {text}\n\nКласифікуйте настрій у "
        "документі. Відповідайте {labels_str}, і нічого більше.",
    ),
}



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn<1.6.0",
# ]
# ///

"""Create the Czech HellaSwag dataset and upload to HF Hub."""

from collections import Counter

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split

from .constants import (
    CHOICES_MAPPING,
    MAX_NUM_CHARS_IN_INSTRUCTION,
    MAX_NUM_CHARS_IN_OPTION,
    MAX_REPETITIONS,
    MIN_NUM_CHARS_IN_INSTRUCTION,
    MIN_NUM_CHARS_IN_OPTION,
)


def main() -> None:
    """Create the Czech HellaSwag dataset and upload to HF Hub."""
    dataset_id = "CZLC/cs_hellaswag"

    # Load the dataset
    dataset = load_dataset(dataset_id, split="test")
    assert isinstance(dataset, Dataset)

    # Convert the dataset to a pandas DataFrame
    df = dataset.to_pandas()
    assert isinstance(df, pd.DataFrame)

    # Remove the samples with overly short or long texts
    df = df[
        (df["query"].str.len() >= MIN_NUM_CHARS_IN_INSTRUCTION)  # Check text
        & (df["query"].str.len() <= MAX_NUM_CHARS_IN_INSTRUCTION)  # Check text
        & df.choices.map(
            lambda choices: min(len(choice) for choice in choices)
            >= MIN_NUM_CHARS_IN_OPTION
            and max(len(choice) for choice in choices) <= MAX_NUM_CHARS_IN_OPTION
        )
    ]

    def is_repetitive(text: str) -> bool:
        """Return True if the text is repetitive."""
        max_repetitions = max(Counter(text.split()).values())
        return max_repetitions > MAX_REPETITIONS

    # Remove overly repetitive samples
    df = df.loc[
        ~df.query.apply(is_repetitive)
        & ~df.choices.map(
            lambda choices: any(is_repetitive(choice) for choice in choices)
        )
    ]

    # Make a `text` column with all the options in it
    df["text"] = [
        row.query.replace("\n", " ").strip() + "\n"
        f"{CHOICES_MAPPING['cs']}:\n"
        "a. " + row.choices[0].replace("\n", " ").strip() + "\n"
        "b. " + row.choices[1].replace("\n", " ").strip() + "\n"
        "c. " + row.choices[2].replace("\n", " ").strip() + "\n"
        "d. " + row.choices[3].replace("\n", " ").strip()
        for _, row in df.iterrows()
    ]

    # Map the `gold` index to a letter for the `label` column
    df["label"] = df["gold"].apply(lambda x: chr(97 + x))

    # Only keep the columns `text` and `label`
    df = df[["text", "label"]]

    # Create train, validation, and test splits
    train_size = 1024
    val_size = 256
    test_size = 2048

    # First, split off the test set
    train_val_df, test_df = train_test_split(df, test_size=test_size, random_state=4242)

    # Then, split the remaining data into train and validation sets
    train_df, val_df = train_test_split(
        train_val_df, test_size=val_size, random_state=4242
    )

    # Ensure the train set is exactly 1024 samples
    train_df = train_df.sample(n=train_size, random_state=4242)

    # Reset the index
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Define the dataset ID for the Hugging Face Hub
    dataset_id = "EuroEval/hellaswag-cs-mini"

    # Push the dataset to the Hugging Face Hub
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



"""All CATALAN dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import CATALAN
from ..tasks import COMMON_SENSE, KNOW, LA, NER, RC, SENT, SUMM

### Official datasets ###

GUIA_CAT_CONFIG = DatasetConfig(
    name="guia-cat",
    pretty_name="GuiaCat",
    source="EuroEval/guia-cat-mini",
    task=SENT,
    languages=[CATALAN],
)

SCALA_CA_CONFIG = DatasetConfig(
    name="scala-ca",
    pretty_name="ScaLA-ca",
    source="EuroEval/scala-ca",
    task=LA,
    languages=[CATALAN],
)

WIKIANN_CA_CONFIG = DatasetConfig(
    name="wikiann-ca",
    pretty_name="WikiANN-ca",
    source="EuroEval/wikiann-ca-mini",
    task=NER,
    languages=[CATALAN],
)

MULTI_WIKI_QA_CA_CONFIG = DatasetConfig(
    name="multi-wiki-qa-ca",
    pretty_name="MultiWikiQA-ca",
    source="EuroEval/multi-wiki-qa-ca-mini",
    task=RC,
    languages=[CATALAN],
)

DACSA_CA_CONFIG = DatasetConfig(
    name="dacsa-ca",
    pretty_name="DACSA-ca",
    source="EuroEval/dacsa-ca-mini",
    task=SUMM,
    languages=[CATALAN],
)

MMLU_CA_CONFIG = DatasetConfig(
    name="mmlu-ca",
    pretty_name="MMLU-ca",
    source="EuroEval/mmlu-ca-mini",
    task=KNOW,
    languages=[CATALAN],
)

WINOGRANDE_CA_CONFIG = DatasetConfig(
    name="winogrande-ca",
    pretty_name="Winogrande-ca",
    source="EuroEval/winogrande-ca",
    task=COMMON_SENSE,
    languages=[CATALAN],
    _labels=["a", "b"],
)



"""All Lithuanian dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import LITHUANIAN
from ..tasks import COMMON_SENSE, KNOW, LA, NER, RC, SENT, SUMM

### Official datasets ###

ATSILIEPIMAI_CONFIG = DatasetConfig(
    name="atsiliepimai",
    pretty_name="Atsiliepimai",
    source="EuroEval/atsiliepimai",
    task=SENT,
    languages=[LITHUANIAN],
)

SCALA_LT_CONFIG = DatasetConfig(
    name="scala-lt",
    pretty_name="ScaLA-lt",
    source="EuroEval/scala-lt",
    task=LA,
    languages=[LITHUANIAN],
)

WIKIANN_LT_CONFIG = DatasetConfig(
    name="wikiann-lt",
    pretty_name="WikiANN-lt",
    source="EuroEval/wikiann-lt-mini",
    task=NER,
    languages=[LITHUANIAN],
)

MULTI_WIKI_QA_LT_CONFIG = DatasetConfig(
    name="multi-wiki-qa-lt",
    pretty_name="MultiWikiQA-lt",
    source="EuroEval/multi-wiki-qa-lt-mini",
    task=RC,
    languages=[LITHUANIAN],
)

LRYTAS_CONFIG = DatasetConfig(
    name="lrytas",
    pretty_name="Lrytas",
    source="EuroEval/lrytas-mini",
    task=SUMM,
    languages=[LITHUANIAN],
)

LT_HISTORY_CONFIG = DatasetConfig(
    name="lt-history",
    pretty_name="LT-History",
    source="EuroEval/lt-history",
    task=KNOW,
    languages=[LITHUANIAN],
)

WINOGRANDE_LT_CONFIG = DatasetConfig(
    name="winogrande-lt",
    pretty_name="Winogrande-lt",
    source="EuroEval/winogrande-lt",
    task=COMMON_SENSE,
    languages=[LITHUANIAN],
    _labels=["a", "b"],
)

### Unofficial datasets ###

LITHUANIAN_EMOTIONS_CONFIG = DatasetConfig(
    name="lithuanian-emotions",
    pretty_name="Lithuanian Emotions",
    source="EuroEval/lithuanian-emotions-mini",
    task=SENT,
    languages=[LITHUANIAN],
    unofficial=True,
)



"""All Portuguese dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import EUROPEAN_PORTUGUESE, PORTUGUESE
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, KNOW, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

SST2_PT_CONFIG = DatasetConfig(
    name="sst2-pt",
    pretty_name="SST2-pt",
    source="EuroEval/sst2-pt-mini",
    task=SENT,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
    _labels=["positive", "negative"],
)

SCALA_PT = DatasetConfig(
    name="scala-pt",
    pretty_name="ScaLA-pt",
    source="EuroEval/scala-pt",
    task=LA,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

HAREM_CONFIG = DatasetConfig(
    name="harem",
    pretty_name="HAREM",
    source="EuroEval/harem",
    task=NER,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

MULTI_WIKI_QA_PT_CONFIG = DatasetConfig(
    name="multi-wiki-qa-pt",
    pretty_name="MultiWikiQA-pt",
    source="EuroEval/multi-wiki-qa-pt-pt-mini",
    task=RC,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

PUBLICO_CONFIG = DatasetConfig(
    name="publico",
    pretty_name="Publico",
    source="EuroEval/publico-mini",
    task=SUMM,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

MMLU_PT_CONFIG = DatasetConfig(
    name="mmlu-pt",
    pretty_name="MMLU-pt",
    source="EuroEval/mmlu-pt-mini",
    task=KNOW,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

GOLDENSWAG_PT_CONFIG = DatasetConfig(
    name="goldenswag-pt",
    pretty_name="GoldenSwag-pt",
    source="EuroEval/goldenswag-pt-mini",
    task=COMMON_SENSE,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
)

VALEU_PT_CONFIG = DatasetConfig(
    name="valeu-pt",
    pretty_name="VaLEU-pt",
    source="EuroEval/european-values-pt",
    task=EUROPEAN_VALUES,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
    splits=["test"],
    bootstrap_samples=False,
    _instruction_prompt="{text}",
)


### Unofficial datasets ###

BOOLQ_PT_CONFIG = DatasetConfig(
    name="boolq-pt",
    pretty_name="BoolQ-pt",
    source="EuroEval/boolq-pt",
    task=MCRC,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
    unofficial=True,
)

WINOGRANDE_PT_CONFIG = DatasetConfig(
    name="winogrande-pt",
    pretty_name="Winogrande-pt",
    source="EuroEval/winogrande-pt",
    task=COMMON_SENSE,
    languages=[PORTUGUESE, EUROPEAN_PORTUGUESE],
    _labels=["a", "b"],
    unofficial=True,
)



"""All Danish dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import DANISH
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, KNOW, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

ANGRY_TWEETS_CONFIG = DatasetConfig(
    name="angry-tweets",
    pretty_name="AngryTweets",
    source="EuroEval/angry-tweets-mini",
    task=SENT,
    languages=[DANISH],
)

SCALA_DA_CONFIG = DatasetConfig(
    name="scala-da",
    pretty_name="ScaLA-da",
    source="EuroEval/scala-da",
    task=LA,
    languages=[DANISH],
)

DANSK_CONFIG = DatasetConfig(
    name="dansk",
    pretty_name="DANSK",
    source="EuroEval/dansk-mini",
    task=NER,
    languages=[DANISH],
)

MULTI_WIKI_QA_DA_CONFIG = DatasetConfig(
    name="multi-wiki-qa-da",
    pretty_name="MultiWikiQA-da",
    source="EuroEval/multi-wiki-qa-da-mini",
    task=RC,
    languages=[DANISH],
)

NORDJYLLAND_NEWS_CONFIG = DatasetConfig(
    name="nordjylland-news",
    pretty_name="Nordjylland News",
    source="EuroEval/nordjylland-news-mini",
    task=SUMM,
    languages=[DANISH],
)

DANSKE_TALEMAADER_CONFIG = DatasetConfig(
    name="danske-talemaader",
    pretty_name="Danske Talemåder",
    source="EuroEval/danske-talemaader",
    task=KNOW,
    languages=[DANISH],
)

DANISH_CITIZEN_TESTS_CONFIG = DatasetConfig(
    name="danish-citizen-tests",
    pretty_name="Danish Citizen Tests",
    source="EuroEval/danish-citizen-tests-updated",
    task=KNOW,
    languages=[DANISH],
)

HELLASWAG_DA_CONFIG = DatasetConfig(
    name="hellaswag-da",
    pretty_name="HellaSwag-da",
    source="EuroEval/hellaswag-da-mini",
    task=COMMON_SENSE,
    languages=[DANISH],
)

VALEU_DA_CONFIG = DatasetConfig(
    name="valeu-da",
    pretty_name="ValEU-da",
    source="EuroEval/european-values-da",
    task=EUROPEAN_VALUES,
    languages=[DANISH],
    splits=["test"],
    bootstrap_samples=False,
)


### Unofficial datasets ###

DANE_CONFIG = DatasetConfig(
    name="dane",
    pretty_name="DaNE",
    source="EuroEval/dane-mini",
    task=NER,
    languages=[DANISH],
    unofficial=True,
)

MMLU_DA_CONFIG = DatasetConfig(
    name="mmlu-da",
    pretty_name="MMLU-da",
    source="EuroEval/mmlu-da-mini",
    task=KNOW,
    languages=[DANISH],
    unofficial=True,
)

ARC_DA_CONFIG = DatasetConfig(
    name="arc-da",
    pretty_name="ARC-da",
    source="EuroEval/arc-da-mini",
    task=KNOW,
    languages=[DANISH],
    unofficial=True,
)

BELEBELE_DA_CONFIG = DatasetConfig(
    name="belebele-da",
    pretty_name="Belebele-da",
    source="EuroEval/belebele-da-mini",
    task=MCRC,
    languages=[DANISH],
    unofficial=True,
)

SCANDIQA_DA_CONFIG = DatasetConfig(
    name="scandiqa-da",
    pretty_name="ScandiQA-da",
    source="EuroEval/scandiqa-da-mini",
    task=RC,
    languages=[DANISH],
    unofficial=True,
)

GOLDENSWAG_DA_CONFIG = DatasetConfig(
    name="goldenswag-da",
    pretty_name="GoldenSwag-da",
    source="EuroEval/goldenswag-da-mini",
    task=COMMON_SENSE,
    languages=[DANISH],
    unofficial=True,
)

WINOGRANDE_DA_CONFIG = DatasetConfig(
    name="winogrande-da",
    pretty_name="Winogrande-da",
    source="EuroEval/winogrande-da",
    task=COMMON_SENSE,
    languages=[DANISH],
    _labels=["a", "b"],
    unofficial=True,
)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the Lithuanian summarisation dataset Lrytas."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi

from .constants import MAX_NUM_CHARS_IN_ARTICLE, MIN_NUM_CHARS_IN_ARTICLE


def main() -> None:
    """Create the Lithuanian Lrytas summarisation mini dataset and upload to HF Hub."""
    dataset_id = "alexandrainst/lrytas-summarization"

    dataset = load_dataset(dataset_id, token=True)
    assert isinstance(dataset, DatasetDict)

    # The dataset has splits: train, val, test
    # Each sample has: url, title, summary, text
    # We want: text = "{title}\n\n{text}", target_text = summary

    def make_columns(sample: dict) -> dict:
        """Map the dataset to have the text and target_text columns.

        Args:
            sample: A sample from the dataset.

        Returns:
            A sample with the text and target_text columns.
        """
        sample["text"] = f"{sample['title']}\n\n{sample['text']}"
        sample["target_text"] = sample["summary"]
        return sample

    dataset = dataset.map(make_columns)
    df = dataset["train"].to_pandas()
    assert isinstance(df, pd.DataFrame)

    # Only work with samples where the text is not very large or small
    lengths = df.text.str.len()
    lower_bound = MIN_NUM_CHARS_IN_ARTICLE
    upper_bound = MAX_NUM_CHARS_IN_ARTICLE
    df = df[lengths.between(lower_bound, upper_bound)]
    df = df.reset_index(drop=True)

    # Create validation split
    val_size = 256
    val_df = df.sample(n=val_size, random_state=4242)
    remaining = df.drop(index=val_df.index.tolist())
    val_df = val_df.reset_index(drop=True)

    # Create test split
    test_size = 2048
    test_df = remaining.sample(n=test_size, random_state=4242)
    remaining2 = remaining.drop(index=test_df.index.tolist())
    test_df = test_df.reset_index(drop=True)

    # Create train split
    train_size = 1024
    train_df = remaining2.sample(n=train_size, random_state=4242)
    train_df = train_df.reset_index(drop=True)

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Collect datasets in a dataset dictionary
    mini_dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    mini_dataset_id = "EuroEval/lrytas-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(mini_dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    mini_dataset.push_to_hub(mini_dataset_id, private=True)


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "beautifulsoup4==4.14.2",
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "joblib==1.4.2",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "tqdm==4.67.1",
# ]
# ///

"""Create the HotterAndColderSentiment dataset and upload it to the HF Hub."""

import datetime as dt
import io
import logging
import re
import warnings
from time import sleep
from zipfile import ZipFile

import joblib
import pandas as pd
import requests as rq
from bs4 import BeautifulSoup, NavigableString, Tag  #  type: ignore[missing-import]
from datasets import Dataset, DatasetDict, Split
from huggingface_hub import HfApi
from tqdm.auto import tqdm

from .constants import MAX_NUM_CHARS_IN_DOCUMENT, MIN_NUM_CHARS_IN_DOCUMENT  # noqa

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("create_hotter_and_colder")


def main() -> None:
    """Create the HotterAndColderSentiment dataset and upload it to the HF Hub.

    Raises:
        HTTPError:
            If the dataset could not be downloaded.
    """
    # Download the ZIP file
    url = (
        "https://repository.clarin.is/repository/xmlui/bitstream/handle/"
        "20.500.12537/352/Icelandic_Sentiment_Corpus.zip"
    )
    response = rq.get(url=url)
    response.raise_for_status()

    # Get the unhydrated data from the ZIP file
    with ZipFile(file=io.BytesIO(initial_bytes=response.content)) as zip_file:
        csv_files = [
            zip_file.read(name=file_name)
            for file_name in zip_file.namelist()
            if file_name.endswith(".csv")
        ]
        assert len(csv_files) == 1, (
            f"Expected one CSV file in the ZIP file, but found {len(csv_files)}."
        )
        df = pd.read_csv(filepath_or_buffer=io.BytesIO(initial_bytes=csv_files[0]))

    # Set up the dataframe for the dataset, resulting in 'text' and 'label' columns,
    # with 'text' being the hydrated comment content and 'label' being the sentiment
    # label ('negative', 'neutral', 'positive')
    df = hydrate_data(df=df.query("annotation_task_name == 'sentiment'"))
    df = df.rename(columns=dict(comment_content="text", label_given_by_user="label"))
    df.label = df.label.map({"0": "negative", "1": "neutral", "2": "positive"})
    df = df.dropna(subset="label")
    df = df.drop(columns=[col for col in df.columns if col not in ["text", "label"]])

    # Create validation split
    val_size = 256
    val_df = df.sample(n=val_size, random_state=4242)
    df = df.drop(index=val_df.index.tolist())

    # Create train split
    train_size = 1024
    train_df = df.sample(n=train_size, random_state=4242)

    # Create test split, just being the rest of the samples
    test_df = df.drop(index=train_df.index.tolist())

    # Sampling maintains the original index, so we need to reset it, as otherwise the
    # conversion to Dataset objects will create extra index columns
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Only work with samples where the document is not very large or small
    # We do it after we have made the splits to ensure that the dataset is minimally
    # affected.
    new_train_df = train_df.copy()
    new_train_df["text_len"] = new_train_df.text.str.len()
    new_train_df = new_train_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_val_df = val_df.copy()
    new_val_df["text_len"] = new_val_df.text.str.len()
    new_val_df = new_val_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_test_df = test_df.copy()
    new_test_df["text_len"] = new_test_df.text.str.len()
    new_test_df = new_test_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )

    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(new_train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(new_val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(new_test_df, split=Split.TEST),
        }
    )

    dataset_id = "EuroEval/hotter-and-colder-sentiment"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    dataset.push_to_hub(dataset_id, private=True)


def hydrate_data(df: pd.DataFrame) -> pd.DataFrame:
    """Hydrate the data in the DataFrame.

    Args:
        df:
            The DataFrame containing the data to hydrate.

    Returns:
        The DataFrame with the hydrated data.
    """
    with warnings.catch_warnings():
        warnings.simplefilter(
            action="ignore", category=pd.errors.SettingWithCopyWarning
        )
        df.comment_datetime = pd.to_datetime(arg=df.comment_datetime)

    with joblib.Parallel(n_jobs=-1, backend="threading") as parallel:
        comment_texts = parallel(
            joblib.delayed(scrape_comment_content)(
                url=row.full_link, target_datetime=row.comment_datetime
            )
            for _, row in tqdm(df.iterrows(), total=len(df), desc="Hydrating comments")
        )

    with warnings.catch_warnings():
        warnings.simplefilter(
            action="ignore", category=pd.errors.SettingWithCopyWarning
        )
        df["comment_content"] = comment_texts

    return df.dropna(subset=["comment_content"])


def scrape_comment_content(url: str, target_datetime: dt.datetime) -> str | None:
    """Scrape the content of a comment with a given datetime from a URL.

    Args:
        url:
            The URL to scrape the content from.
        target_datetime:
            The datetime of the comment to find.

    Returns:
        The content of the comment with the given datetime, or None if the comment could
        not be found.
    """
    num_attempts = 10
    for _ in range(num_attempts):
        try:
            response = rq.get(url=url)
            response.raise_for_status()
            break
        except (rq.HTTPError, rq.ConnectionError):
            sleep(1)
            continue
    else:
        logger.error(f"Could not fetch {url!r} after {num_attempts} attempts.")
        return None

    soup = BeautifulSoup(markup=response.content, features="html.parser")

    comment_divs: list[Tag] = soup.find_all(
        name="div",
        attrs={
            "class": [
                "comment even registered",
                "comment odd registered",
                "comment even registered own",
                "comment odd registered own",
                "comment odd unregistered",
                "comment even unregistered",
            ]
        },
    )
    for comment in comment_divs:
        comment_body = comment.find(name="div", class_="comment-body")
        signature_body = comment.find(name="p", class_="comment-signature")

        # Sanity check to ensure that we're not leaving any comments behind; this
        # ensures that the two bodies are either bs4 Tag objects or None.
        assert not isinstance(comment_body, NavigableString)
        assert not isinstance(signature_body, NavigableString)

        if isinstance(comment_body, Tag) and isinstance(signature_body, Tag):
            comment_datetime = extract_datetime_from_signature(
                signature_body=signature_body
            )
            if comment_datetime == target_datetime:
                return comment_body.text.strip()
    else:
        logger.error(f"Error scraping comment content for {url!r}")
        return None


def extract_datetime_from_signature(signature_body: Tag) -> dt.datetime | None:
    """Parse the signature to extract the datetime information.

    Args:
        signature_body:
            The signature body to extract the datetime information from.

    Returns:
        The extracted datetime information, or None if no datetime information was
        found.
    """
    signature_str = "".join([str(child) for child in signature_body.children]).strip()
    signature_str = re.sub("\n", " ", signature_str)
    signature_str = re.sub(r"\s+", " ", signature_str)

    datetime_match = re.search(
        pattern=r"(\d{1,2}\.\d{1,2}\.\d{4})\s+kl\.\s+(\d{1,2}:\d{2})",
        string=signature_str,
    )
    if not datetime_match:
        return None

    date_str, time_str = datetime_match.groups()
    return dt.datetime.strptime(f"{date_str} {time_str}", "%d.%m.%Y %H:%M")


if __name__ == "__main__":
    main()
